import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class Waveguide_Arc(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the waveguide arc.

    Author: Mustafa Hammood     mustafa@siepic.com
    """

    def __init__(self):
        # Important: initialize the super class
        super(Waveguide_Arc, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param("radius", self.TypeDouble, "Radius", default=10)
        self.param("wg_width", self.TypeDouble, "Waveguide Width", default=0.5)
        self.param("start_angle", self.TypeDouble, "Start Angle", default=0)
        self.param("stop_angle", self.TypeDouble, "Stop Angle", default=45)
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        # hidden parameters, can be used to query this component:
        self.param(
            "p1",
            self.TypeShape,
            "DPoint location of pin1",
            default=Point(-10000, 0),
            hidden=True,
            readonly=True,
        )
        self.param(
            "p2",
            self.TypeShape,
            "DPoint location of pin2",
            default=Point(0, 10000),
            hidden=True,
            readonly=True,
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Waveguide_Arc(R=" + ("%.3f" % self.radius) + ")"

    def produce(self, layout, layers, parameters, cell):
        """
        coerce parameters (make consistent)
        """
        self._layers = layers
        self.cell = cell
        self._param_values = parameters
        self.layout = layout

        # cell: layout cell to place the layout
        # LayerSiN: which layer to use
        # r: radius
        # w: waveguide width
        # start_angle: starting angle of the arc
        # stop_agnle: stopping angle of the arc
        # length units in dbu
        import math
        from SiEPIC.utils import arc_wg

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout

        LayerSi = self.silayer
        LayerSiN = self.silayer_layer
        #    LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        from SiEPIC.extend import to_itype

        w = to_itype(self.wg_width, dbu)
        r = to_itype(self.radius, dbu)
        start_angle = self.start_angle
        stop_angle = self.stop_angle

        if start_angle > stop_angle:
            start_angle = self.stop_angle
            stop_angle = self.start_angle

        deg_to_rad = math.pi / 180.0

        # draw the arc
        x = 0
        y = 0

        from SiEPIC._globals import PIN_LENGTH as pin_length

        self.cell.shapes(LayerSiN).insert(arc_wg(r, w, start_angle, stop_angle))
        # Create the pins, as short paths:

        # Pin on the right side:
        x = r * math.cos(start_angle * deg_to_rad)
        y = r * math.sin(start_angle * deg_to_rad)

        x_pin = math.cos((90 - start_angle) * deg_to_rad) * pin_length / 2
        y_pin = math.sin((90 - start_angle) * deg_to_rad) * pin_length / 2

        p2 = [Point(x - x_pin, y + y_pin), Point(x + x_pin, y - y_pin)]
        p2c = Point(x, y)
        self.set_p2 = p2c
        self.p2 = p2c
        pin = Path(p2, w)
        self.cell.shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, x, y)
        text = Text("pin2", t)
        shape = self.cell.shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Pin on the left side:
        x = round(r * math.cos(stop_angle * deg_to_rad))
        y = round(r * math.sin(stop_angle * deg_to_rad))

        x_pin = math.cos((90.0 - stop_angle) * deg_to_rad) * pin_length / 2
        y_pin = math.sin((90.0 - stop_angle) * deg_to_rad) * pin_length / 2

        p1 = [Point(x + x_pin, y - y_pin), Point(x - x_pin, y + y_pin)]
        p1c = Point(x, y)
        self.set_p1 = p1c
        self.p1 = p1c
        pin = Path(p1, w)
        self.cell.shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, x, y)
        text = Text("pin1", t)
        shape = self.cell.shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        x = 0
        y = 0
        # layout_arc_wg_dbu(self.cell, LayerDevRecN, x, y, r, w*3, start_angle, stop_angle)
        self.cell.shapes(LayerDevRecN).insert(arc_wg(r, w * 3, start_angle, stop_angle))



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class Contra_DC_Bent(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the bent-CDC half ring resonator pcell

    Author: Mustafa Hammood     mustafa@siepic.com
    """

    def __init__(self):
        # Important: initialize the super class
        super(Contra_DC_Bent, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param(
            "silayer_gratings",
            self.TypeLayer,
            "Si Gratings Layer",
            default=TECHNOLOGY["Si"],
        )
        self.param("radius", self.TypeDouble, "Radius (um)", default=25)
        self.param("gap", self.TypeDouble, "Gap (um)", default=0.28)
        self.param("bus_width", self.TypeDouble, "Bus Width (um)", default=0.45)
        self.param("ring_width", self.TypeDouble, "Ring Width (um)", default=0.55)

        self.param("period", self.TypeDouble, "Gratings Period (nm)", default=318)
        self.param(
            "deltaWB", self.TypeDouble, "Bus Corrugation Width (um)", default=0.04
        )
        self.param(
            "deltaWR", self.TypeDouble, "Ring Corrugation Width (um)", default=0.05
        )
        self.param("gamma", self.TypeDouble, "N (number of corrugations)", default=135)

        self.param("busBend", self.TypeDouble, "Bus-to-straight bend raidus", default=5)

        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        # hidden parameters, can be used to query this component:
        self.param(
            "p1",
            self.TypeShape,
            "DPoint location of pin1",
            default=Point(-10000, 0),
            hidden=True,
            readonly=True,
        )
        self.param(
            "p2",
            self.TypeShape,
            "DPoint location of pin2",
            default=Point(0, 10000),
            hidden=True,
            readonly=True,
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Contra_DC_Bent_(R=" + ("%.3f" % self.radius) + ")"

    def produce(self, layout, layers, parameters, cell):
        """
        coerce parameters (make consistent)
        """
        self._layers = layers
        self.cell = cell
        self._param_values = parameters
        self.layout = layout

        # cell: layout cell to place the layout
        # LayerSiN: which layer to use
        # r: radius
        # w: waveguide width
        # length units in dbu

        from math import pi, cos, sin
        from SiEPIC.utils import arc_wg_xy

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout

        LayerSi = self.silayer
        LayerSiN_gratings = self.silayer_gratings_layer
        LayerSiN = self.silayer_layer
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        from SiEPIC.extend import to_itype

        w_bus = to_itype(self.bus_width, dbu)
        w_ring = to_itype(self.ring_width, dbu)
        w = w_ring
        r = int(round(self.radius / dbu))
        gap = int(round(self.gap / dbu))
        period = self.period
        deltaWR = int(round(self.deltaWR / dbu))
        N = int(self.gamma)
        deltaWB = int(round(self.deltaWB / dbu))
        busBendR = int(round(self.busBend / dbu))

        # Center of everything
        x = 0
        y = 0

        # Angle of CDC portion, also bend angle!
        periodAngle = (180 / pi) * (period / 2) / (r + w_ring / 2 + gap / 2)

        # Bend angle
        bendAngle = (180 / pi) * (N * period / 2) / (r + w_ring / 2 + gap / 2)

        # Bend radius of bus
        rBus = r + gap + w_bus / 2 + w_ring / 2

        N_input = N
        # Normalized number of corrugations to periodAngle
        N = N * periodAngle * 2

        # Convention:   Set A is the CDC portion that is not offsetted, set B is offsetted by corrugation width

        # Ring CDCs (set A)
        ii = periodAngle * 2
        while ii < N + periodAngle * 1.5:
            self.cell.shapes(LayerSiN_gratings).insert(
                arc_wg_xy(
                    x,
                    y,
                    r - deltaWR / 2,
                    w_ring,
                    90 + bendAngle - ii,
                    90 + bendAngle - ii - periodAngle,
                )
            )
            ii = ii + periodAngle
            ii = ii + periodAngle

        # Ring CDCs (set B)
        ii = periodAngle
        while ii < N:
            self.cell.shapes(LayerSiN_gratings).insert(
                arc_wg_xy(
                    x,
                    y,
                    r + deltaWR / 2,
                    w_ring,
                    90 + bendAngle - ii,
                    90 + bendAngle - ii - periodAngle,
                )
            )
            ii = ii + periodAngle
            ii = ii + periodAngle

        # Bus CDCs (set A)
        ii = periodAngle
        while ii < N:
            self.cell.shapes(LayerSiN_gratings).insert(
                arc_wg_xy(
                    x,
                    y,
                    rBus - deltaWB / 2,
                    w_bus,
                    90 + bendAngle - ii,
                    90 + bendAngle - ii - periodAngle,
                )
            )
            ii = ii + periodAngle
            ii = ii + periodAngle

        # Bus CDCs (set B)
        ii = periodAngle * 2
        while ii < N + periodAngle * 1.5:
            self.cell.shapes(LayerSiN_gratings).insert(
                arc_wg_xy(
                    x,
                    y,
                    rBus + deltaWB / 2,
                    w_bus,
                    90 + bendAngle - ii,
                    90 + bendAngle - ii - periodAngle,
                )
            )
            ii = ii + periodAngle
            ii = ii + periodAngle

        # Ring non-CDC left
        self.cell.shapes(LayerSiN).insert(
            arc_wg_xy(x, y, r, w_ring, 180 - (90 - bendAngle), 180)
        )

        # Ring non-CDC right
        self.cell.shapes(LayerSiN).insert(arc_wg_xy(x, y, r, w_ring, 0, 90 - bendAngle))

        bendAngleRad = (pi / 180) * bendAngle
        # Draw Bus non-CDC Waveguide
        y_center = rBus * cos(bendAngleRad) + busBendR * cos(bendAngleRad)
        x_center = rBus * sin(bendAngleRad) + busBendR * sin(bendAngleRad)

        self.cell.shapes(LayerSiN).insert(
            arc_wg_xy(x_center, y_center, busBendR, w_bus, 270 - bendAngle, 270)
        )
        self.cell.shapes(LayerSiN).insert(
            arc_wg_xy(-x_center, y_center, busBendR, w_bus, 270, 270 - bendAngle)
        )

        # Create the pins, as short paths:
        from SiEPIC._globals import PIN_LENGTH as pin_length

        # Pin on the right side:
        y_center = rBus * cos(bendAngleRad) + busBendR * cos(bendAngleRad) - busBendR
        p2 = [
            Point(-pin_length / 2 + x_center, y_center),
            Point(pin_length / 2 + x_center, y_center),
        ]
        p2c = Point(x_center, y_center)
        self.set_p2 = p2c
        self.p2 = p2c
        pin = Path(p2, w_bus)
        self.cell.shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, x_center, y_center)
        text = Text("pin2", t)
        shape = self.cell.shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Pin on the left side:
        x_center = -x_center
        p1 = [
            Point(pin_length / 2 + x_center, y_center),
            Point(-pin_length / 2 + x_center, y_center),
        ]
        p1c = Point(x_center, y_center)
        self.set_p1 = p1c
        self.p1 = p1c
        pin = Path(p1, w_bus)
        self.cell.shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, x_center, y_center)
        text = Text("pin1", t)
        shape = self.cell.shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Pin on the bottom left side:
        p3 = [Point(x - r, pin_length / 2 + y), Point(x - r, -pin_length / 2 + y)]
        p3c = Point(x - r, y)
        self.set_p3 = p3c
        self.p3 = p3c
        pin = Path(p3, w_ring)
        self.cell.shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, x - r, y)
        text = Text("pin3", t)
        shape = self.cell.shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Pin on the bottom right side:
        p4 = [Point(x + r, y + pin_length / 2), Point(x + r, y - pin_length / 2)]
        p4c = Point(x + r, y)
        self.set_p4 = p4c
        self.p4 = p4c
        pin = Path(p4, w_ring)
        self.cell.shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, x + r, y)
        text = Text("pin4", t)
        shape = self.cell.shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        self.cell.shapes(LayerDevRecN).insert(arc_wg_xy(x, y, r + gap, w * 5, 0, 180))



import pya
import math
from pya import *
from SiEPIC.utils import get_technology_by_name


class ebeam_PWB_surface_taper(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the photonic wirebond surface taper
    Author: Becky Lin
    V1: July 30, 2020
    V2: Sept 8, 2020
    V3: Dec 4, 2020
    blin@ece.ubc.ca

    Lukas, 2024/02/12: rounding error on width conversion to integer, fixed using to_itype.
    """

    def __init__(self):
        # Important: initialize the super class
        super(ebeam_PWB_surface_taper, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        self.param("Wtip", self.TypeDouble, "Width of tip (microns)", default=0.13)
        self.param(
            "Wwaveguide", self.TypeDouble, "width of waveguide (microns)", default=0.5
        )
        self.param(
            "edge_offset",
            self.TypeDouble,
            "tip to edge gap (microns, recommend min 30um)",
            default=30,
        )
        self.param(
            "taper_length",
            self.TypeDouble,
            "length of taper (microns, min 45um; max 100um)",
            default=65,
        )
        self.param(
            "OxOp_angle", self.TypeInt, "Oxide open interface angle (int)", default=8
        )
        self.param(
            "show_deepTrench",
            self.TypeBoolean,
            "Show deep trench layer? !Min DT and OxOp is 5um",
            default=True,
        )
        self.param(
            "ZEP_Process",
            self.TypeBoolean,
            "Fabrication using UBC ZEP process",
            default=False,
        )
        self.param(
            "ZEP_Trench", self.TypeDouble, "Trench width for ZEP process", default=10
        )
        self.param(
            "Grouse_Run",
            self.TypeBoolean,
            "Fabrication using ANT Grouse Run",
            default=False,
        )

        # declare the layers
        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param(
            "oxopen",
            self.TypeLayer,
            "OxOpen Layer",
            default=TECHNOLOGY["Oxide open (to BOX)"],
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])
        self.param(
            "fibertarget",
            self.TypeLayer,
            "Fiber Target Layer",
            default=TECHNOLOGY["FbrTgt"],
        )
        self.param(
            "ZEP_invert",
            self.TypeLayer,
            "31 Si for ZEP inverse",
            default=LayerInfo(31, 0),
        )
        self.param(
            "Grouse_Si_Keepout",
            self.TypeLayer,
            "Grouse Si Keep-out Layer",
            default=LayerInfo(63, 0),
        )

    def can_create_from_shape_impl(self):
        return False

    def produce(self, layout, layers, parameters, cell):
        # This is the main part of the implementation: create the layout

        #
        self.cell = cell
        self._param_values = parameters
        self.layout = layout

        # import neccessary libraries
        from math import pi, sin, cos, tan
        from SiEPIC._globals import PIN_LENGTH as pin_length
        from SiEPIC.extend import to_itype

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSiN = ly.layer(self.silayer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerOxOpN = ly.layer(self.oxopen)
        # LayerDTN = ly.layer(self.deeptrench)
        LayerTEXTN = ly.layer(self.textl)
        LayerFbrTgtN = ly.layer(self.fibertarget)
        Layer31SiN = ly.layer(self.ZEP_invert)
        Layer63N = ly.layer(self.Grouse_Si_Keepout)

        w1 = to_itype(self.Wtip, dbu)
        w2 = to_itype(self.Wwaveguide, dbu)
        offset = self.edge_offset
        angle = self.OxOp_angle * (pi / 180)
        excl = 750
        notile = 15
        y_oxop = 20
        l = self.taper_length
        wT = self.ZEP_Trench

        # draw deeptrench layer (to indicate edge)
        pts6 = [
            Point(0, excl / dbu),
            Point(-100 / dbu, excl / dbu),
            Point(-100 / dbu, -excl / dbu),
            Point(0, -excl / dbu),
        ]

        # place deeptrench layer
        OxOp_DT_gap = 0

        if self.show_deepTrench:
            # shapes(LayerDTN).insert(Polygon(pts6))
            OxOp_DT_gap = 5

            # draw no surface element window for edge coupling
            pts5 = [
                Point(0, excl / dbu),
                Point((offset + l + excl) / dbu, excl / dbu),
                Point((offset + l + excl) / dbu, -excl / dbu),
                Point(0, -excl / dbu),
            ]

            t = Trans(Trans.R0, 0, (excl - 800) / dbu)
            text = Text("edge of chip", t)
            shape = shapes(LayerTEXTN).insert(text)
            shape.text_size = 7 / dbu

        else:
            # draw no surface element window for on chip wirebonding
            pts5 = [
                Point(-(excl + l + 150) / dbu, excl / dbu),
                Point((offset + l + excl) / dbu, excl / dbu),
                Point((offset + l + excl) / dbu, -excl / dbu),
                Point(-(excl + l + 150) / dbu, -excl / dbu),
            ]

        # draw the taper
        pts = [
            Point(offset / dbu, w1 / 2),
            Point((l + offset) / dbu, w2 / 2),
            Point((l + offset) / dbu, -w2 / 2),
            Point(offset / dbu, -w1 / 2),
        ]
        # pts = [Point(offset/dbu,-0.001/dbu),Point((offset-20)/dbu,-0.001/dbu),Point((offset-20)/dbu,0.001/dbu),Point(offset/dbu,0.001/dbu),Point(offset/dbu,w1/2),Point((l+offset)/dbu,w2/2),Point((l+offset)/dbu,-w2/2),Point(offset/dbu,-w1/2)]

        # place the taper
        shapes(LayerSiN).insert(Polygon(pts))

        # draw waveguide section
        Lw2 = 15
        pts7 = [
            Point((offset + l) / dbu, (w2 / 2)),
            Point((offset + l + Lw2) / dbu, (w2 / 2)),
            Point((offset + l + Lw2) / dbu, (-w2 / 2)),
            Point((offset + l) / dbu, (-w2 / 2)),
        ]

        # place waveguide section
        shapes(LayerSiN).insert(Polygon(pts7))

        # draw alignment marker
        pts1 = [
            Point((0 + offset) / dbu, 12.4 / dbu),
            Point((0 + offset) / dbu, 9 / dbu),
            Point((3.5 + offset) / dbu, 9 / dbu),
            Point((3.5 + offset) / dbu, 12.5 / dbu),
            Point((0.1 + offset) / dbu, 12.5 / dbu),
            Point((0 + offset) / dbu, 12.6 / dbu),
            Point((0 + offset) / dbu, 16 / dbu),
            Point((-3.5 + offset) / dbu, 16 / dbu),
            Point((-3.5 + offset) / dbu, 12.5 / dbu),
            Point((-0.1 + offset) / dbu, 12.5 / dbu),
        ]

        pts2 = [
            Point((0 + 50 + offset) / dbu, 12.4 / dbu),
            Point((0 + 50 + offset) / dbu, 9 / dbu),
            Point((3.5 + 50 + offset) / dbu, 9 / dbu),
            Point((3.5 + 50 + offset) / dbu, 12.5 / dbu),
            Point((0.1 + 50 + offset) / dbu, 12.5 / dbu),
            Point((0 + 50 + offset) / dbu, 12.6 / dbu),
            Point((0 + 50 + offset) / dbu, 16 / dbu),
            Point((-3.5 + 50 + offset) / dbu, 16 / dbu),
            Point((-3.5 + 50 + offset) / dbu, 12.5 / dbu),
            Point((-0.1 + 50 + offset) / dbu, 12.5 / dbu),
        ]

        # place alignment marker
        shapes(LayerSiN).insert(Polygon(pts1))
        shapes(LayerSiN).insert(Polygon(pts2))

        # draw oxide open window
        dy_oxop = 5
        dx = dy_oxop * tan(angle)
        ddx = sin(angle)
        ddy = cos(angle)
        # pts3 = [Point(OxOp_DT_gap/dbu,y_oxop/dbu),Point((offset+l+10+dx-1)/dbu,dy_oxop/dbu),Point((offset+l+10+dx-ddx)/dbu,(y_oxop-ddy)/dbu),Point((offset+l+10-dx)/dbu,-(y_oxop-10)/dbu),Point(OxOp_DT_gap/dbu,-(y_oxop-10)/dbu)]

        pts3 = [
            Point(OxOp_DT_gap / dbu, y_oxop / dbu),
            Point((offset + l + 10 + dx - 1) / dbu, y_oxop / dbu),
            Point((offset + l + 10 + dx - 1) / dbu, dy_oxop / dbu),
            Point((offset + l + 10 - dx - 1) / dbu, -dy_oxop / dbu),
            Point((offset + l + 10 - dx - 1) / dbu, -(y_oxop - 10) / dbu),
            Point(OxOp_DT_gap / dbu, -(y_oxop - 10) / dbu),
        ]

        # place oxide open window
        shapes(LayerOxOpN).insert(Polygon(pts3))

        # draw devrec box
        pts4 = [
            Point(0, y_oxop / dbu),
            Point((offset + l + 10 + dx) / dbu, y_oxop / dbu),
            Point((offset + l + 10 + dx) / dbu, -(y_oxop - 10) / dbu),
            Point(0, -(y_oxop - 10) / dbu),
        ]

        # place devrec box
        shapes(LayerDevRecN).insert(Polygon(pts5))

        # EXCLUSION LAYERS################################################################

        # place no surface element window
        # shapes(LayerTEXTN).insert(Polygon(pts5))

        # place no surface element text warning
        t = Trans(Trans.R0, 0, (excl - 730) / dbu)
        text = Text(
            "Recommend no surface elements \n(e.g. metals pads, grating couplers, oxide open, etc) \nwithin this DevRec box.\nSurface elements in this box may be covered by cladding photoresist",
            t,
        )
        shape = shapes(LayerTEXTN).insert(text)
        shape.text_size = 7 / dbu

        # ZEP PROCESS#####################################################################

        # draw ZEP process, expanded taper trench used during layer inverting
        # path1 = Path([Point(offset,w2/2),Point((l+offset),w2/2),Point((l+offset),-w2/2),Point(offset,-w2/2)],1)
        # pts8 = [Point(offset/dbu,w2/2),Point((l+offset)/dbu,w2/2),Point((l+offset)/dbu,-w2/2),Point(offset/dbu,-w2/2)]

        # if self.ZEP_Process == True:
        # shapes(Layer31SiN).insert(Polygon(pts8))
        # x = (offset + l)/dbu
        # t = Trans(Trans.R0, x, 0)
        # path1_t = pin.transformed(t)
        # shapes(LayerSiN).insert(path1_t)

        ZEPTrench_pts = [
            Point((offset + 65) / dbu, -wT / dbu),
            Point((offset - 25) / dbu, -wT / dbu),
            Point((offset - 25) / dbu, wT / dbu),
            Point((offset + 65) / dbu, wT / dbu),
        ]
        if self.ZEP_Process == True:
            # place the extended Si Keepout at the tip of taper
            shapes(Layer63N).insert(Polygon(ZEPTrench_pts))

        # GROUSE SI KEEPOUT ##############################################################

        # draw the extended Si Keepout at the tip of taper
        grouseRun_pts = [
            Point(offset / dbu, -2.175 / dbu),
            Point((offset - 25) / dbu, -2.175 / dbu),
            Point((offset - 25) / dbu, 2.175 / dbu),
            Point(offset / dbu, 2.175 / dbu),
        ]
        if self.Grouse_Run == True:
            # place the extended Si Keepout at the tip of taper
            shapes(Layer63N).insert(Polygon(grouseRun_pts))

        #################################################################################

        # draw and place pin on the waveguide:
        x = (offset + l + Lw2) / dbu
        t = Trans(Trans.R0, x, 0)
        pin = Path([Point(-pin_length / 2, 0), Point(pin_length / 2, 0)], w2)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        def circle(x, y, r):
            npts = 180
            theta = 2 * math.pi / npts
            pts = []
            for i in range(0, npts):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            (x + r * math.cos(i * theta)) / 1,
                            (y + r * math.sin(i * theta)) / 1,
                        )
                    )
                )
            return pts

        # draw fibre target circle
        align_circle = circle(offset / dbu, 0, 2 / dbu)

        # place fibre target circle
        shapes(LayerFbrTgtN).insert(Polygon(align_circle))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        if self.show_deepTrench:
            return (
                "PWB_taper (tip_width="
                + ("%.2f" % self.Wtip)
                + ",waveguide_width="
                + ("%.2f" % self.Wwaveguide)
                + ",taper_length="
                + ("%.1f" % self.taper_length)
                + ",Angle="
                + ("%s" % self.OxOp_angle)
                + ",edge_offset="
                + ("%.2f" % self.edge_offset)
                + ")"
            )
        else:
            return (
                "PWB_taper (tip_width="
                + ("%.2f" % self.Wtip)
                + ",waveguide_width="
                + ("%.2f" % self.Wwaveguide)
                + ",taper_length="
                + ("%.1f" % self.taper_length)
                + ",Angle="
                + ("%s" % self.OxOp_angle)
                + ")"
            )



from pya import *
import pya


class Waveguide_SBend(pya.PCellDeclarationHelper):
    """
    Input:
    """

    def __init__(self):
        # Important: initialize the super class
        super(Waveguide_SBend, self).__init__()
        from SiEPIC.utils import get_technology_by_name

        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("length", self.TypeDouble, "Waveguide length", default=10.0)
        self.param("height", self.TypeDouble, "Waveguide offset height", default=2)
        self.param(
            "wg_width", self.TypeDouble, "Waveguide width (microns)", default=0.5
        )
        #    self.param("radius", self.TypeDouble, "Waveguide bend radius (microns)", default = 5)
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Waveguide_SBend_%s-%.3f" % (self.length, self.wg_width)

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        from SiEPIC.utils.layout import layout_waveguide_sbend_bezier
        from SiEPIC.extend import to_itype

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        length = to_itype(self.length, dbu)
        w = to_itype(self.wg_width, dbu)
        h = to_itype(self.height, dbu)

        #    waveguide_length = layout_waveguide_sbend(self.cell, LayerSiN, pya.Trans(Trans.R0, 0,0), w, r, h, length)
        waveguide_length = to_itype(
            layout_waveguide_sbend_bezier(
                self.cell,
                LayerSiN,
                Trans(),
                self.wg_width,
                self.wg_width,
                self.height,
                self.length,
            ),
            dbu,
        )

        from SiEPIC._globals import PIN_LENGTH as pin_length

        # Pins on the waveguide:
        x = length
        t = Trans(Trans.R0, x, h)
        pin = Path([Point(-pin_length / 2, 0), Point(pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = to_itype(0.4, dbu)
        shape.text_halign = 2

        x = 0
        t = Trans(Trans.R0, x, 0)
        pin = Path([Point(pin_length / 2, 0), Point(-pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = to_itype(0.4, dbu)

        # Compact model information
        t = Trans(Trans.R0, 0, 0)
        text = Text("Lumerical_INTERCONNECT_library=Design kits/EBeam", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = to_itype(0.1, dbu)
        t = Trans(Trans.R0, 0, w * 2)
        text = Text("Component=ebeam_wg_integral_1550", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = to_itype(0.1, dbu)
        t = Trans(Trans.R0, 0, -w * 2)
        text = Text(
            "Spice_param:wg_length=%.3fu wg_width=%.3fu"
            % (waveguide_length * dbu, self.wg_width),
            t,
        )
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = to_itype(0.1, dbu)
        #    t = Trans(Trans.R0, 0, -w*3)
        #    text = Text ('Extra length = %.4fu, Shortest length = %.4fu' % (straight_l*dbu, (length-2*straight_l)*dbu), t )
        #    shape = shapes(LayerDevRecN).insert(text)
        #    shape.text_size = 0.1/dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        box1 = Box(0, min(-w * 3, h - w * 3), length, max(w * 3, h + w * 3))
        shapes(LayerDevRecN).insert(box1)


#    print('2020/11/28 update')



# $description: Unit testing: all EBeam libraries' cells
# $show-in-menu
# $group-name: Examples_EBeam
# $menu-path: siepic_menu.exlayout.end
# Unit testing for all library layout fixed cells and PCells

tech_name = "EBeam"

import os
from SiEPIC._globals import Python_Env

print("KLayout running in mode: %s" % Python_Env)
from SiEPIC.utils.layout import new_layout
from SiEPIC.scripts import instantiate_all_library_cells


def test_all_library_cells():
    if Python_Env == "Script":
        # Load the PDK from a folder, e.g, GitHub, when running externally from the KLayout Application
        import sys

        path = os.path.dirname(os.path.realpath(__file__))
        sys.path.insert(0, os.path.abspath(os.path.join(path, "../../..")))
        import siepic_ebeam_pdk

    # Create a new layout
    topcell, ly = new_layout(tech_name, "UnitTesting", GUI=True, overwrite=True)

    # Instantiate all cells
    instantiate_all_library_cells(topcell)

    """
    # Save the layout
    path_out = os.path.dirname(os.path.realpath(__file__))
    ly.write(os.path.join(path_out,'a.gds'))
    export_layout(topcell, path_out,'a_static',format='oas')
    print('done')
    """

    # Check if there are any errors
    for cell_id in topcell.each_child_cell():
        c = ly.cell(cell_id)
        error_shapes = c.shapes(ly.error_layer())
        for error in error_shapes.each():
            raise Exception("Error in cell: %s, %s" % (c.name, error.text))
        if c.is_empty() or c.bbox().area() == 0:
            raise Exception("Empty cell: %s" % c.name)


if __name__ == "__main__":
    test_all_library_cells()



# This PCell instantiates one ebeam_dc_halfring_straight and one waveguide

import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class ring_singlebus(pya.PCellDeclarationHelper):
    """
    The PCell uses one ebeam_dc_halfring_straight PCell and one waveguide
    """

    def __init__(self):
        # Important: initialize the super class
        super(ring_singlebus, self).__init__()

        from SiEPIC.utils import load_Waveguides_by_Tech

        self.technology_name = "EBeam"
        TECHNOLOGY = get_technology_by_name(self.technology_name)

        # declare the parameters
        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param("r", self.TypeDouble, "Radius", default=10)
        self.param("w", self.TypeDouble, "Waveguide Width", default=0.5)
        self.param("g", self.TypeDouble, "Gap", default=0.2)
        self.param("Lc", self.TypeDouble, "Coupler Length", default=0.0)
        self.param(
            "orthogonal_identifier",
            self.TypeInt,
            "Orthogonal identifier (1=TE, 2=TM)",
            default=1,
        )
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])

        # Load all waveguides types
        self.waveguide_types = load_Waveguides_by_Tech(self.technology_name)
        p = self.param(
            "waveguide_type",
            self.TypeList,
            "Waveguide Type",
            default=self.waveguide_types[0]["name"],
        )
        for wa in self.waveguide_types:
            p.add_choice(wa["name"], wa["name"])

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return (
            "ring_singlebus(R="
            + ("%.3f" % self.r)
            + ",g="
            + ("%g" % (1000 * self.g))
            + ",Lc="
            + ("%g" % (1000 * self.Lc))
            + ",orthogonal_identifier="
            + ("%s" % self.orthogonal_identifier)
            + ")"
        )

    def can_create_from_shape_impl(self):
        return False

    def produce_impl(self):
        # This is the main part of the implementation: create the layout

        # print(self.layout.technology_name)  # the technology is not assigned in a PCell, for some reason
        self.layout.technology_name = self.technology_name

        # Create the PCell
        pcell = self.layout.create_cell(
            "ebeam_dc_halfring_straight",
            self.technology_name,
            {
                "silayer": self.silayer,
                "r": self.r,
                "w": self.w,
                "g": self.g,
                "Lc": self.Lc,
                "orthogonal_identifier": self.orthogonal_identifier,
                "pinrec": self.pinrec,
                "devrec": self.devrec,
                "textl": self.textl,
            },
        )
        if pcell == None:
            raise Exception(
                "problem! cannot create PCell from library: %s" % self.technology_name
            )
        # Instantiate the PCell in the layout
        inst = self.cell.insert(
            pya.CellInstArray(pcell.cell_index(), pya.Trans(pya.Trans.R0, 0, 0))
        )

        path = pya.DPath(
            [
                pya.DPoint(-self.r, self.r + self.g + self.w),
                pya.DPoint(-self.r, self.r * 2 + self.g + self.w),
                pya.DPoint(self.r, self.r * 2 + self.g + self.w),
                pya.DPoint(self.r, self.r + self.g + self.w),
            ],
            0.5,
        )

        wg_pcell = self.layout.create_cell(
            "Waveguide",
            self.technology_name,
            {"waveguide_type": self.waveguide_type, "path": path},
        )
        if wg_pcell == None:
            raise Exception(
                "problem! cannot create Waveguide from library: %s"
                % self.technology_name
            )
        inst = self.cell.insert(
            pya.CellInstArray(wg_pcell.cell_index(), pya.Trans(pya.Trans.R0, 0, 0))
        )

        # print("Done drawing the layout for - ring_singlebus: %.3f-%g" % ( self.r, self.g) )



# $description: Contra-directional coupler designs (EBeam)
# $show-in-menu
# $group-name: Examples_EBeam
# $menu-path: siepic_menu.exlayout.begin


"""

Create a complete layout for several contra-directional coupler circuits

using KLayout & SiEPIC-Tools scripted layout

usage:
 - the SiEPIC EBeam Library
 - Uncomment path_to_waveguide to convert everything to waveguides, or just do it on your own from SiEPIC Tools menu.
 - Run circuit simulations either using OPICS or INTERCONNECT

Script will generate an series of devices with the selected parameters, and generate GCs for testing with automated measurement labels

Author: Lukas Chrostowski, Mustafa Hammood

March 2024
"""

print("SiEPIC_EBeam_PDK: Example - Example - Contra-directional coupler designs.py")

# Import KLayout-Python API
import pya
from pya import *

import SiEPIC
from SiEPIC._globals import Python_Env
from SiEPIC.scripts import zoom_out, export_layout
import os

if Python_Env == "Script":
    try:
        # For external Python mode, when installed using pip install siepic_ebeam_pdk
        import siepic_ebeam_pdk
    except:
        # Load the PDK from a folder, e.g, GitHub, when running externally from the KLayout Application
        import os, sys

        path_GitHub = os.path.expanduser("~/Documents/GitHub/")
        sys.path.append(os.path.join(path_GitHub, "SiEPIC_EBeam_PDK/klayout"))
        import siepic_ebeam_pdk

tech_name = "EBeam"

# Import from SiEPIC-Tools
from SiEPIC.extend import to_itype
from SiEPIC.scripts import connect_pins_with_waveguide, connect_cell

from packaging import version

if version.parse(SiEPIC.__version__) < version.parse("0.5.1"):
    pya.MessageBox.warning(
        "Errors",
        "This example requires SiEPIC-Tools version 0.5.1 or greater.",
        pya.MessageBox.Ok,
    )


# define layout parameters in the class below
# ALL distance parameters are in microns unless specified otherwise
class parameters:
    """Define the circuit layout parameters"""

    # Default contraDC PCell parameters
    component_contraDC = "contra_directional_coupler"
    libname = "EBeam"
    N = 1000
    period = 0.316
    g = 100
    w1 = 560
    w2 = 440
    dW1 = 0.05
    dW2 = 0.025
    sine = 0
    a = 2.7

    # routing and device placement parameters
    x_offset = 170  # spacing between grating couplers columns
    wg_bend_radius = 35  # waveguide routes bend radius
    device_spacing = 7.3  # spacing between devices
    wg_pitch = (
        5  # spacing between waveguides ( keep > 2 microns to minimize cross coupling)
    )
    pol = "TE"
    waveguide_type = "Strip TE 1550 nm, w=500 nm"
    gc_pitch = 127  # spacing between grating couplers array (for a single device)
    cdc_offset = 20
    name = "contraDC"


params = parameters()


def import_fixed_cells(ly):
    """Import the fixed cells from the library, and add them to the layout"""
    params.cell_gc = ly.create_cell(
        "ebeam_gc_%s1550" % params.pol.lower(), "EBeam"
    )  # .cell_index()
    # params.cell_y = ly.create_cell("ebeam_y_1550", "EBeam") # .cell_index()

    params.gc_length = 41  # Length of a grating coupler cell
    params.gc_height = 30  # Height of a grating coupler cell


def ebeam_c_te_mux_1ch_standard_1543nm(
    topcell,
    x_pos=0,
    y_pos=0,
    N=params.N,
    period=params.period,
    g=0.1,
    w1=0.56,
    w2=0.44,
    dW1=0.048,
    dW2=0.024,
    sine=0,
    a=2.7,
    wg_width=0.5,
    pol=params.pol,
):
    """Create a layout of a contradirectional coupler, with grating couplers
    This is a known-good design by Mustafa Hammood,
    having been fabricated by Applied Nanotools, and tested at UBC
    It is in the C-band and is used for the Process Control Monitor (PCM) structure (C-PCM).
    Works well when cascaded.
    Strip waveguides
    Bandwidth ~6 nm;	Wavelength 1543 nm
    Data: single stage, cascaded, exists on every PCM
    function by Lukas Chrostowski
    """

    # Create a sub-cell for our contraDC layout
    cell = topcell.layout().create_cell("contraDC_GCarray")

    # place the cell in the top cell
    t = Trans(Trans.R0, x_pos, y_pos)
    topcell.insert(CellInstArray(cell.cell_index(), t))
    ly = topcell.layout()

    # Grating couplers, Ports 1, 2, 3, 4 (top-down)
    instGCs = []
    for i in range(4):
        t = Trans(
            Trans.R0,
            to_itype(params.gc_length, ly.dbu),
            to_itype(params.gc_height / 2 + i * params.gc_pitch, ly.dbu),
        )
        instGCs.append(cell.insert(CellInstArray(params.cell_gc.cell_index(), t)))
    t = Trans(
        Trans.R0,
        to_itype(params.gc_length, ly.dbu),
        to_itype(params.gc_height / 2 + params.gc_pitch * 2, ly.dbu),
    )
    text = Text(
        "opt_in_%s_1550_device_%s%dN%dperiod%dg%dwa%dwb%ddwa%ddwb%dsine%sa"
        % (
            pol,
            params.name,
            N,
            1000 * period,
            1000 * g,
            1000 * w1,
            1000 * w2,
            1000 * dW1,
            1000 * dW2,
            sine,
            a,
        ),
        t,
    )
    # text = Text ("opt_in_TE_1550_device_contraDC1"
    shape = cell.shapes(ly.layer(ly.TECHNOLOGY["Text"])).insert(text)
    shape.text_size = 1.5 / ly.dbu

    # contraDC PCell
    pcell = ly.create_cell(
        params.component_contraDC,
        params.libname,
        {
            "sbend": 1,
            "number_of_periods": N,
            "grating_period": period,
            "gap": g,
            "wg1_width": w1,
            "wg2_width": w2,
            "corrugation_width1": dW1,
            "corrugation_width2": dW2,
            "sinusoidal": sine,
            "index": a,
        },
    )
    if not pcell:
        raise Exception(
            "Cannot find cell %s in library %s."
            % (params.component_contraDC, params.libname)
        )
    t = Trans(
        Trans.R90,
        to_itype(params.gc_length + params.cdc_offset, ly.dbu),
        to_itype(params.gc_height / 2 + params.gc_pitch * 0.2, ly.dbu),
    )
    instCDC = cell.insert(CellInstArray(pcell.cell_index(), t))

    # Waveguides:
    connect_pins_with_waveguide(
        instGCs[3], "opt1", instCDC, "opt3", waveguide_type=params.waveguide_type
    )
    connect_pins_with_waveguide(
        instGCs[2],
        "opt1",
        instCDC,
        "opt4",
        waveguide_type=params.waveguide_type,
        turtle_A=[5, 90, 30, -90],
        turtle_B=[5, 90],
    )
    connect_pins_with_waveguide(
        instGCs[1],
        "opt1",
        instCDC,
        "opt2",
        waveguide_type=params.waveguide_type,
        turtle_A=[5, -90, 30, 90],
        turtle_B=[5, -90],
    )
    connect_pins_with_waveguide(
        instGCs[0], "opt1", instCDC, "opt1", waveguide_type=params.waveguide_type
    )

    return cell


def layout_contraDC_circuits(newlayout=True):
    """
    Generates contraDC circuits.
    Either create a new layout using the EBeam technology,
        newlayout = True
    or delete everything in the present layout
        newlayout = False
    """

    from SiEPIC.utils.layout import new_layout, floorplan
    from SiEPIC.utils import select_paths, get_layout_variables

    """
    Create a new layout using the EBeam technology, with a top cell
    """
    topcell, ly = new_layout(
        tech_name, "SiEPIC_EBeam_contraDC_circuits", GUI=True, overwrite=not newlayout
    )

    # Import the grating couplers, and other fixed cells
    import_fixed_cells(ly)

    # create a floor plan
    # 605 x 410 microns is the space allocation for the edX course and openEBL
    # https://siepic.ca/openebl/
    floorplan(topcell, 605e3, 410e3)

    # Create the contraDC circuits
    ebeam_c_te_mux_1ch_standard_1543nm(topcell)

    # Zoom out
    zoom_out(topcell)

    # Save
    path = os.path.dirname(os.path.realpath(__file__))
    filename = "openEBL_Contra-directional-coupler-design"
    file_out = export_layout(
        topcell, path, filename, relative_path="..", format="oas", screenshot=False
    )

    from SiEPIC.verification import layout_check

    print("SiEPIC_EBeam_PDK: - verification")

    file_lyrdb = os.path.join(path, filename + ".lyrdb")
    num_errors = layout_check(
        cell=topcell, verbose=False, GUI=True, file_rdb=file_lyrdb
    )

    if Python_Env == "Script":
        from SiEPIC.utils import klive

        klive.show(file_out, lyrdb_filename=file_lyrdb, technology=tech_name)

    if num_errors == 0:
        print("Functional verification: Passed with 0 errors")
    else:
        print("Functional verification: Failed with %s errors" % num_errors)


layout_contraDC_circuits(newlayout=False)


# All done!



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class ebeam_irph_wg(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the IRPH straight waveguide.

    Authors: Jaspreet Jhoja, Lukas Chrostowski
    """

    def __init__(self):
        # Important: initialize the super class
        super(ebeam_irph_wg, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("w", self.TypeDouble, "Waveguide Width", default=0.5)
        self.param("w_rib", self.TypeDouble, "Waveguide Rib Width", default=24)
        self.param("length", self.TypeDouble, "Waveguide length", default=100)
        self.param("n_w", self.TypeDouble, "N Width", default=22)
        self.param("npp_w", self.TypeDouble, "N++ Width (um)", default=10)
        self.param("npp_dw", self.TypeDouble, "N++ Edge to Si Edge (um)", default=2)
        self.param("vc_dw", self.TypeDouble, "VC Edge to Si Edge (um)", default=5)
        self.param("vc_w", self.TypeDouble, "VC Width", default=5)
        self.param("m_dw", self.TypeDouble, "Metal edge to Si Edge (um)", default=4)
        self.param("m_w", self.TypeDouble, "Metal Width", default=10)
        self.param(
            "overlay",
            self.TypeDouble,
            "Overlay accuracy (optical litho) (um)",
            default=1,
        )
        self.param(
            "overlay_ebl", self.TypeDouble, "Overlay accuracy (EBL) (um)", default=0.05
        )
        self.param(
            "io_wg_type",
            self.TypeBoolean,
            "I/O waveguide type, 0 - strip; 1 - rib",
            default=0,
        )

        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param(
            "siriblayer",
            self.TypeLayer,
            "Si rib Layer",
            default=TECHNOLOGY["Si - 90 nm rib"],
        )
        self.param("nlayer", self.TypeLayer, "N Layer", default=TECHNOLOGY["Si N"])
        self.param(
            "npplayer", self.TypeLayer, "N++ Layer", default=TECHNOLOGY["Si N++"]
        )
        self.param("vclayer", self.TypeLayer, "VC Layer", default=TECHNOLOGY["VC"])
        self.param(
            "mlayer", self.TypeLayer, "Metal Layer", default=TECHNOLOGY["M2_router"]
        )
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=LayerInfo(10, 0))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "ebeam_irph_wg"  # (R=" + ('%.3f' % self.r) + ",g=" + ('%g' % (1000*self.g)) + ")"

    def can_create_from_shape_impl(self):
        return False

    def produce_impl(self):
        # This is the main part of the implementation: create the layout

        from SiEPIC._globals import PIN_LENGTH
        from SiEPIC.extend import to_itype

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSiN = ly.layer(self.silayer)
        LayerSiRibN = ly.layer(self.siriblayer)
        LayerNN = ly.layer(self.nlayer)
        LayerNPPN = ly.layer(self.npplayer)
        LayerVCN = ly.layer(self.vclayer)
        LayerMN = ly.layer(self.mlayer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        TextLayerN = ly.layer(self.textl)

        w = to_itype(self.w, dbu)
        w_rib = to_itype(self.w_rib, dbu)
        length = to_itype(self.length, dbu)
        n_w = to_itype(self.n_w, dbu)
        npp_w = to_itype(self.npp_w, dbu)
        npp_dw = to_itype(self.npp_dw, dbu)
        vc_w = to_itype(self.vc_w, dbu)
        vc_dw = to_itype(self.vc_dw, dbu)
        m_dw = to_itype(self.m_dw, dbu)
        m_w = to_itype(self.m_w, dbu)
        overlay = to_itype(self.overlay, dbu)
        overlay_ebl = to_itype(self.overlay_ebl, dbu)

        if self.io_wg_type == 0:
            ## define strip to slab transfer taper parameters
            N = 100  # Number of points for the input/output slab taper
            order = 3  # input/output slab taper curve
            in_taper = to_itype(5, dbu)  # length of taper
            taper_w = to_itype(2, dbu)
            # add input slab taper
            pts = []
            for i in range(0, N + 1):
                pts.append(
                    Point(
                        -in_taper + in_taper / N * i,
                        (w - overlay_ebl * 2) / 2
                        + ((taper_w - (w - overlay_ebl * 2)) / (N**order)) * (i**order),
                    )
                )
            for i in range(0, N + 1):
                pts.append(
                    Point(
                        -in_taper + in_taper / N * (N - i),
                        -(w - overlay_ebl * 2) / 2
                        - ((taper_w - (w - overlay_ebl * 2)) / (N**order))
                        * ((N - i) ** order),
                    )
                )
            self.cell.shapes(LayerSiRibN).insert(Polygon(pts))
            # add output slab taper
            pts = []
            for i in range(0, N + 1):
                pts.append(
                    Point(
                        length + in_taper - in_taper / N * i,
                        (w - overlay_ebl * 2) / 2
                        + ((taper_w - (w - overlay_ebl * 2)) / (N**order)) * (i**order),
                    )
                )
            for i in range(0, N + 1):
                pts.append(
                    Point(
                        length + in_taper - in_taper / N * (N - i),
                        -(w - overlay_ebl * 2) / 2
                        - ((taper_w - (w - overlay_ebl * 2)) / (N**order))
                        * ((N - i) ** order),
                    )
                )
            self.cell.shapes(LayerSiRibN).insert(Polygon(pts))
        else:
            in_taper = 0  # length of taper
            taper_w = 0  # width of taper

        # draw the waveguide
        xtop = 0 - in_taper
        ytop = -1 * (w / 2)
        xbottom = length + in_taper
        ybottom = 1 * (w / 2)
        wg1 = Box(xtop, -w / 2, xbottom, w / 2)
        shapes(LayerSiN).insert(wg1)

        # Pins on the bottom waveguide side:
        pin = Path(
            [Point(xtop + PIN_LENGTH / 2, 0), Point(xtop - PIN_LENGTH / 2, 0)], w
        )
        shapes(LayerPinRecN).insert(pin)
        text = Text("pin1", Trans(Trans.R0, xtop, 0))
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        pin = Path(
            [Point(xbottom - PIN_LENGTH / 2, 0), Point(xbottom + PIN_LENGTH / 2, 0)], w
        )
        shapes(LayerPinRecN).insert(pin)
        text = Text("pin2", Trans(Trans.R0, xbottom, 0))
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # draw N
        b = Box(0, -n_w / 2, length, n_w / 2)
        shapes(LayerNN).insert(b)

        # draw NPP
        b = Box(0, npp_dw + w / 2, length, npp_dw + w / 2 + npp_w)
        shapes(LayerNPPN).insert(b)
        b = Box(0, -npp_dw - w / 2, length, -npp_dw - w / 2 - npp_w)
        shapes(LayerNPPN).insert(b)

        # draw via
        b = Box(overlay, vc_dw + w / 2, length - overlay, vc_dw + w / 2 + vc_w)
        shapes(LayerVCN).insert(b)
        b = Box(overlay, -vc_dw - w / 2, length - overlay, -vc_dw - w / 2 - vc_w)
        shapes(LayerVCN).insert(b)

        # draw metal
        b = Box(0, m_dw + w / 2, length, m_dw + w / 2 + m_w)
        shapes(LayerMN).insert(b)
        b = Box(0, -m_dw - w / 2, length, -m_dw - w / 2 - m_w)
        shapes(LayerMN).insert(b)

        # devrec layer
        w_devrec = 2 * max(
            m_dw + m_w + w / 2, vc_dw + w / 2 + vc_w, npp_dw + w / 2 + npp_w, w_rib / 2
        )
        pts = []
        pts = [
            Point(0, -w_devrec / 2 - w),
            Point(0, -taper_w),
            Point(0 - in_taper, -taper_w),
            Point(0 - in_taper, +taper_w),
            Point(0, +taper_w),
            Point(0, w_devrec / 2 + w),
            Point(length, w_devrec / 2 + w),
            Point(length, +taper_w),
            Point(length + in_taper, +taper_w),
            Point(length + in_taper, -taper_w),
            Point(length, -taper_w),
            Point(length, -w_devrec / 2 - w),
        ]
        self.cell.shapes(LayerDevRecN).insert(Polygon(pts))

        # Si rib layer
        b = Box(0, w_rib / 2, length, -w_rib / 2)
        shapes(LayerSiRibN).insert(b)

        t = Trans(Trans.R0, 0, 0)
        text = Text("Component=ebeam_irph_wg", t)
        shape = shapes(LayerDevRecN).insert(text)
        # shape.text_size = self.r*0.07/dbu


#    print("Done drawing the layout for - ebeam_irph_wg: %.3f-%g" % ( self.r, self.g))



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class Contra_DC_custom(pya.PCellDeclarationHelper):
    """
    Author:   Mustafa Hammood
              Mustafa@siepic.com
    """

    def __init__(self):
        # Important: initialize the super class
        super(Contra_DC_custom, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param(
            "number_of_periods1", self.TypeInt, "Number of grating periods", default=300
        )
        self.param(
            "grating_period1", self.TypeDouble, "Grating period (inner)", default=0.318
        )
        self.param(
            "grating_period2", self.TypeDouble, "Grating period (outer)", default=0.318
        )
        self.param("gap", self.TypeDouble, "Gap (microns)", default=0.15)
        self.param(
            "corrugation_width1outer",
            self.TypeDouble,
            "Waveguide 1 Corrugration width (outer)",
            default=0.03,
        )
        self.param(
            "corrugation_width1inner",
            self.TypeDouble,
            "Waveguide 1 Corrugration width (inner)",
            default=0.04,
        )
        self.param(
            "corrugation_width2outer",
            self.TypeDouble,
            "Waveguide 2 Corrugration width (inner)",
            default=0.01,
        )
        self.param(
            "corrugation_width2inner",
            self.TypeDouble,
            "Waveguide 2 Corrugration width (outer)",
            default=0.01,
        )
        self.param("wg1_width", self.TypeDouble, "Waveguide 1 width", default=0.45)
        self.param("wg2_width", self.TypeDouble, "Waveguide 2 width", default=0.55)
        self.param("index", self.TypeDouble, "Gaussian Index", default=2.8)
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=LayerInfo(10, 0))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Contra_DC_dualBragg%s-%.3f" % (
            self.number_of_periods1,
            self.grating_period1,
        )

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        import math

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        from SiEPIC.extend import to_itype

        # Draw the Bragg grating (bottom):
        width_period1 = to_itype(self.grating_period1 / 2, dbu)
        width_period2 = to_itype(self.grating_period2 / 2, dbu)

        grating_period1 = to_itype(self.grating_period1, dbu)
        grating_period2 = to_itype(self.grating_period2, dbu)

        w1 = to_itype(self.wg1_width, dbu)
        w2 = to_itype(self.wg2_width, dbu)

        GaussianIndex = self.index

        N1 = self.number_of_periods1

        vertical_offset1 = to_itype(self.gap / 2 + self.wg1_width / 2, dbu)
        vertical_offset2 = to_itype(-self.gap / 2 - self.wg2_width / 2, dbu)

        for i in range(0, self.number_of_periods1):
            x1 = to_itype(i * self.grating_period1, dbu)
            x2 = to_itype(i * self.grating_period2, dbu)

            profileFunction = math.exp(
                -0.5 * (2 * GaussianIndex * (i - N1 / 2) / (N1)) ** 2
            )

            profile_outer = (
                to_itype(self.corrugation_width1outer, dbu) * profileFunction
            )
            profile_inner = (
                to_itype(self.corrugation_width1inner, dbu) * profileFunction
            )

            box1outer = Box(
                x1,
                vertical_offset1 + w1 / 2 - profile_outer / 2,
                x1 + width_period1,
                vertical_offset1,
            )
            box2outer = Box(
                x1 + width_period1,
                vertical_offset1 + w1 / 2 + profile_outer / 2,
                x1 + 2 * width_period1,
                vertical_offset1,
            )

            box1inner = Box(
                x2,
                vertical_offset1 - w1 / 2 - profile_inner / 2,
                x2 + width_period2,
                vertical_offset1,
            )
            box2inner = Box(
                x2 + width_period2,
                vertical_offset1 - w1 / 2 + profile_inner / 2,
                x2 + 2 * width_period2,
                vertical_offset1,
            )

            shapes(LayerSiN).insert(box1outer)
            shapes(LayerSiN).insert(box2outer)

            shapes(LayerSiN).insert(box1inner)
            shapes(LayerSiN).insert(box2inner)

            length1 = x1 + grating_period1
            length2 = x2 + grating_period2

        if grating_period1 > grating_period2:
            box_inner = Box(
                length2, vertical_offset1, length1, vertical_offset1 - w1 / 2
            )
            shapes(LayerSiN).insert(box_inner)
            length = x1 + grating_period1
        else:
            box_outer = Box(
                length1, vertical_offset1, length2, vertical_offset1 + w1 / 2
            )
            shapes(LayerSiN).insert(box_outer)
            length = x2 + grating_period2

        # lower waveguide
        for i in range(0, self.number_of_periods1):
            x1 = to_itype(i * self.grating_period1, dbu)
            x2 = to_itype(i * self.grating_period2, dbu)

            profileFunction = math.exp(
                -0.5 * (2 * GaussianIndex * (i - N1 / 2) / (N1)) ** 2
            )

            profile_outer = (
                to_itype(self.corrugation_width2outer, dbu) * profileFunction
            )
            profile_inner = (
                to_itype(self.corrugation_width2inner, dbu) * profileFunction
            )

            box1outer = Box(
                x2,
                vertical_offset2 + w2 / 2 + profile_outer / 2,
                x2 + width_period2,
                vertical_offset2,
            )
            box2outer = Box(
                x2 + width_period2,
                vertical_offset2 + w2 / 2 - profile_outer / 2,
                x2 + 2 * width_period2,
                vertical_offset2,
            )

            box1inner = Box(
                x1,
                vertical_offset2 - w2 / 2 + profile_inner / 2,
                x1 + width_period1,
                vertical_offset2,
            )
            box2inner = Box(
                x1 + width_period1,
                vertical_offset2 - w2 / 2 - profile_inner / 2,
                x1 + 2 * width_period1,
                vertical_offset2,
            )

            shapes(LayerSiN).insert(box1outer)
            shapes(LayerSiN).insert(box2outer)

            shapes(LayerSiN).insert(box1inner)
            shapes(LayerSiN).insert(box2inner)

            length1 = x1 + grating_period1
            length2 = x2 + grating_period2

        if grating_period1 > grating_period2:
            box_inner = Box(
                length2, vertical_offset2, length1, vertical_offset2 + w2 / 2
            )
            shapes(LayerSiN).insert(box_inner)
            length = x1 + grating_period1
        else:
            box_outer = Box(
                length1, vertical_offset2, length2, vertical_offset2 - w2 / 2
            )
            shapes(LayerSiN).insert(box_outer)
            length = x2 + grating_period2

        # Create the pins on the waveguides, as short paths:
        from SiEPIC._globals import PIN_LENGTH as pin_length

        w = to_itype(self.wg2_width, dbu)
        t = Trans(Trans.R0, 0, vertical_offset2)
        pin = Path([Point(pin_length / 2, 0), Point(-pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        w = to_itype(self.wg1_width, dbu)
        t = Trans(Trans.R0, 0, vertical_offset1)
        pin = Path([Point(pin_length / 2, 0), Point(-pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin3", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        w = to_itype(self.wg1_width, dbu)
        t = Trans(Trans.R0, length, vertical_offset1)
        pin = Path([Point(-pin_length / 2, 0), Point(pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        w = to_itype(self.wg2_width, dbu)
        t = Trans(Trans.R0, length, vertical_offset2)
        pin = Path([Point(-pin_length / 2, 0), Point(pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin4", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Compact model information
        t = Trans(Trans.R0, 0, 0)
        text = Text("Lumerical_INTERCONNECT_library=Design kits/ebeam", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu
        t = Trans(Trans.R0, length / 10, 0)
        text = Text("Component=ebeam_contra_dc", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu
        t = Trans(Trans.R0, length / 9, 0)
        text = Text("Spice_param:number_of_periods=%s" % (self.number_of_periods1), t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        t = Trans(Trans.R0, 0, 0)
        path = Path([Point(0, 0), Point(length, 0)], 3 * w)
        shapes(LayerDevRecN).insert(path.simple_polygon())



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name
from SiEPIC.utils import points_per_circle  # ,layout


class swg_fc_test(pya.PCellDeclarationHelper):
    """
    Sub-wavelength-grating fibre coupler PCell litho test structure.

    2017/07/12: Lukas Chrostowski, initial version, based on swg_fc by Tim

    Input:

    """

    def __init__(self):
        # Important: initialize the super class
        super(swg_fc_test, self).__init__()

        # declare the parameters
        self.param(
            "wavelength", self.TypeDouble, "Design Wavelength (micron)", default=2.9
        )
        self.param("n_t", self.TypeDouble, "Fiber Mode", default=1.0)
        self.param("n_e", self.TypeDouble, "Grating Index Parameter", default=3.1)
        self.param("angle_e", self.TypeDouble, "Taper Angle (deg)", default=20.0)
        self.param(
            "grating_length", self.TypeDouble, "Grating Length (micron)", default=3.0
        )
        self.param(
            "taper_length", self.TypeDouble, "Taper Length (micron)", default=32.0
        )
        self.param("dc", self.TypeDouble, "Duty Cycle", default=0.488193)
        self.param("period", self.TypeDouble, "Grating Period", default=1.18939)
        self.param("ff", self.TypeDouble, "Fill Factor", default=0.244319)
        self.param("t", self.TypeDouble, "Waveguide Width (micron)", default=1.0)
        self.param("theta_c", self.TypeDouble, "Insertion Angle (deg)", default=8.0)
        self.param(
            "fab_error", self.TypeDouble, "Fab Process error max (micron)", default=0.05
        )

        # Layer parameters
        TECHNOLOGY = get_technology_by_name("EBeam")
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "swg_fc_test_%.1f-%.2f-%.2f-%.2f-%.2f-%.2f-%.2f-%.2f" % (
            self.wavelength,
            self.theta_c,
            self.period,
            self.dc,
            self.ff,
            self.angle_e,
            self.taper_length,
            self.t,
        )

    #    return "temporary placeholder"

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(self.layer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerTextN = ly.layer(self.textl)

        from math import pi, cos, sin

        lambda_0 = self.wavelength  ##um wavelength of light
        pin_length = 0.5  ##um extra nub for the waveguid attachment

        # Geometry
        wh = self.period * self.dc  ##thick grating
        wl = self.ff * (self.period - wh)  ## thin grating
        spacing = (self.period - wh - wl) / 2  ##space between thick and thin

        gc_number = int(round(self.grating_length / self.period))  ##number of periods
        gc_number = 3
        e = self.n_t * sin((pi / 180) * self.theta_c) / self.n_e
        N = round(
            self.taper_length * (1 + e) * self.n_e / lambda_0
        )  ##allows room for the taper

        start = pi - (pi / 180) * self.angle_e / 2
        stop = pi + (pi / 180) * self.angle_e / 2

        # Draw coupler grating.
        for j in range(gc_number):
            # number of points in the arcs:
            # calculate such that the vertex & edge placement error is < 0.5 nm.
            #   see "SiEPIC_EBeam_functions - points_per_circle" for more details
            radius = N * lambda_0 / (self.n_e * (1 - e)) + j * self.period + spacing
            seg_points = int(
                points_per_circle(radius / dbu) / 360.0 * self.angle_e
            )  # number of points grating arc
            theta_up = []
            for m in range(seg_points + 1):
                theta_up = theta_up + [start + m * (stop - start) / seg_points]
            theta_down = theta_up[::-1]

            ##small one
            r_up = []
            r_down = []
            rng = range(len(theta_up))

            # find the divider to get desired fab error:
            th = min(theta_up)
            div = (2 * sin(th) / self.fab_error) * (
                N * lambda_0 / (self.n_e * (1 - e * cos(th)))
                + j * self.period
                + spacing
            )
            err = (2 * sin(th) / div) * (
                N * lambda_0 / (self.n_e * (1 - e * cos(th)))
                + j * self.period
                + spacing
            )
            #        print("div %s, err (double check) %s" % (div, err))

            for k in rng:
                th = theta_up[k]
                #          print("%s, %s, %s" % (th, sin(th), 1+sin(th)/10.) )
                r_up = r_up + [
                    (1 - sin(th) / div)
                    * (
                        N * lambda_0 / (self.n_e * (1 - e * cos(th)))
                        + j * self.period
                        + spacing
                    )
                ]
            for k in rng[::-1]:
                th = theta_up[k]
                #          print("%s, %s, %s" % (th, sin(th), 1+sin(th)/10.) )
                r_down = r_down + [
                    (1 + sin(th) / div)
                    * (
                        N * lambda_0 / (self.n_e * (1 - e * cos(th)))
                        + j * self.period
                        + spacing
                    )
                ]

            xr = []
            yr = []
            for k in range(len(theta_up)):
                xr = xr + [r_up[k] * cos(theta_up[k])]
                yr = yr + [r_up[k] * sin(theta_up[k])]

            xl = []
            yl = []
            for k in range(len(theta_down)):
                xl = xl + [(r_down[k] + wl) * cos(theta_down[k])]
                yl = yl + [(r_down[k] + wl) * sin(theta_down[k])]

            x = xr + xl
            y = yr + yl

            pts = []
            for i in range(len(x)):
                pts.append(Point.from_dpoint(pya.DPoint(x[i] / dbu, y[i] / dbu)))
            # small_one = core.Boundary(points)

            polygon = Polygon(pts)
            shapes(LayerSiN).insert(polygon)

            if j == 1:
                # text label dimensions, for minor grating:
                # top
                shapes(LayerTextN).insert(
                    Text(
                        "%0.0f" % ((wl + self.fab_error) * 1000),
                        Trans(Trans.R0, xl[0] / dbu, yl[0] / dbu),
                    )
                ).text_size = 0.2 / dbu
                # btm
                shapes(LayerTextN).insert(
                    Text(
                        "%0.0f" % ((wl - self.fab_error) * 1000),
                        Trans(Trans.R0, xl[-1] / dbu, yl[-1] / dbu),
                    )
                ).text_size = 0.2 / dbu
                # mid
                shapes(LayerTextN).insert(
                    Text(
                        "%0.0f" % ((wl) * 1000),
                        Trans(
                            Trans.R0,
                            xl[int(len(theta_up) / 2)] / dbu,
                            yl[int(len(theta_up) / 2)] / dbu,
                        ),
                    )
                ).text_size = 0.2 / dbu

            ##big one
            r_up = []
            r_down = []

            # find the divider to get desired fab error:
            th = min(theta_up)
            div = (2 * sin(th) / self.fab_error) * (
                N * lambda_0 / (self.n_e * (1 - e * cos(th)))
                + j * self.period
                + 2 * spacing
                + wl
            )
            err = (2 * sin(th) / div) * (
                N * lambda_0 / (self.n_e * (1 - e * cos(th)))
                + j * self.period
                + 2 * spacing
                + wl
            )
            #        print("div %s, err (double check) %s" % (div, err))

            rng = range(len(theta_up))
            for k in rng:
                th = theta_up[k]
                r_up = r_up + [
                    (1 - sin(th) / div)
                    * (
                        N * lambda_0 / (self.n_e * (1 - e * cos(th)))
                        + j * self.period
                        + 2 * spacing
                        + wl
                    )
                ]
            for k in rng[::-1]:
                th = theta_up[k]
                r_down = r_down + [
                    (1 + sin(th) / div)
                    * (
                        N * lambda_0 / (self.n_e * (1 - e * cos(th)))
                        + j * self.period
                        + 2 * spacing
                        + wl
                    )
                ]

            xr = []
            yr = []
            for k in range(len(theta_up)):
                xr = xr + [r_up[k] * cos(theta_up[k])]
                yr = yr + [r_up[k] * sin(theta_up[k])]

            xl = []
            yl = []
            for k in range(len(theta_down)):
                xl = xl + [(r_down[k] + wh) * cos(theta_down[k])]
                yl = yl + [(r_down[k] + wh) * sin(theta_down[k])]

            x = xr + xl
            y = yr + yl

            pts = []
            for i in range(len(x)):
                pts.append(Point.from_dpoint(pya.DPoint(x[i] / dbu, y[i] / dbu)))

            polygon = Polygon(pts)
            shapes(LayerSiN).insert(polygon)

            if j == 1:
                # text label dimensions, for major grating:
                # top
                shapes(LayerTextN).insert(
                    Text(
                        "%0.0f" % ((wh + self.fab_error) * 1000),
                        Trans(Trans.R0, xl[0] / dbu, yl[0] / dbu),
                    )
                ).text_size = 0.2 / dbu
                # btm
                shapes(LayerTextN).insert(
                    Text(
                        "%0.0f" % ((wh - self.fab_error) * 1000),
                        Trans(Trans.R0, xl[-1] / dbu, yl[-1] / dbu),
                    )
                ).text_size = 0.2 / dbu
                # mid
                shapes(LayerTextN).insert(
                    Text(
                        "%0.0f" % ((wh) * 1000),
                        Trans(
                            Trans.R0,
                            xl[int(len(theta_up) / 2)] / dbu,
                            yl[int(len(theta_up) / 2)] / dbu,
                        ),
                    )
                ).text_size = 0.2 / dbu



import pya
import math
from pya import *


class ebeam_dream_PWB_edge_couplers_BB(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the ebeam_dream_PWB_edge_couplers_BB

    Authors: Dream Photonics
    """

    def __init__(self):
        # Important: initialize the super class
        super(ebeam_dream_PWB_edge_couplers_BB, self).__init__()
        from SiEPIC.utils import get_technology_by_name

        self.technology_name = "EBeam"
        TECHNOLOGY = get_technology_by_name(self.technology_name)

        # declare the parameters
        self.param(
            "num_channels", self.TypeInt, "Number of Channels (1 - 16)", default=1
        )
        p = self.param(
            "center_wavelength",
            self.TypeList,
            "Center Wavelength of Operation [nm]",
            default="1550",
        )
        p.add_choice("1550 nm", "1550")
        p.add_choice("1310 nm", "1310")

        # declare the layers
        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param(
            "waveguidelayer",
            self.TypeLayer,
            "Waveguide Layer",
            default=TECHNOLOGY["Waveguide"],
        )
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param(
            "fibertarget",
            self.TypeLayer,
            "Fiber Target Layer",
            default=TECHNOLOGY["FbrTgt"],
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])
        self.param("bb", self.TypeLayer, "BB Layer", default=TECHNOLOGY["BlackBox"])

    def can_create_from_shape_impl(self):
        return False

    def produce(self, layout, layers, parameters, cell):
        # This is the main part of the implementation: create the layout

        self.cell = cell
        self._param_values = parameters
        self.layout = layout

        # import neccessary libraries
        from SiEPIC._globals import PIN_LENGTH as pin_length
        from SiEPIC.extend import to_itype

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSiN = ly.layer(self.silayer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerFbrTgtN = ly.layer(self.fibertarget)
        LayerWN = ly.layer(self.waveguidelayer)
        LayerTEXTN = ly.layer(self.textl)
        LayerBBN = ly.layer(self.bb)

        w_tip = to_itype(0.100, dbu)
        l_taper = to_itype(65, dbu)
        num_channels = self.num_channels
        offset = to_itype(0, dbu)
        pitch = to_itype(127, dbu)
        Lw2 = to_itype(15, dbu)
        Lw3 = Lw2 + to_itype(20, dbu)
        wavelength = int(self.center_wavelength)
        waveguide_type = "Strip TE 1550 nm, w=500 nm"

        if num_channels < 1:
            num_channels = 1
        if num_channels > 16:
            num_channels = 16

        if wavelength == 1310:
            waveguide_type = "Strip TE 1310 nm, w=350 nm"
            w_waveguide = to_itype(0.350, dbu)
        elif wavelength == 1550:
            waveguide_type = "Strip TE 1550 nm, w=500 nm"
            w_waveguide = to_itype(0.500, dbu)

        def circle(x, y, r):
            npts = 180
            theta = 2 * math.pi / npts
            pts = []
            for i in range(0, npts):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            (x + r * math.cos(i * theta)) / 1,
                            (y + r * math.sin(i * theta)) / 1,
                        )
                    )
                )
            return pts

        # draw one loopback device
        for ref_loop in range(2):
            # draw fibre target circle
            align_circle = circle(offset, -pitch * (ref_loop + 1), 2 / dbu)

            # place fibre target circle
            shapes(LayerFbrTgtN).insert(Polygon(align_circle))

        # create waveguide to for loopback
        loopback_path = DPath(
            [
                DPoint(0, -127),
                DPoint((offset + l_taper + Lw2) * dbu + 15, -127),
                DPoint((offset + l_taper + Lw2) * dbu + 15, -254),
                DPoint(0, -254),
            ],
            0.5,
        )
        self.layout.technology_name = (
            self.technology_name
        )  # required otherwise "create_cell" doesn't load
        pcell = self.layout.create_cell(
            "Waveguide",
            self.technology_name,
            {"path": loopback_path, "waveguide_type": waveguide_type},
        )
        t = Trans(Trans.R0, 0, 0)
        self.cell.copy(pcell, LayerSiN, LayerBBN)
        wg = self.cell.insert(CellInstArray(pcell.cell_index(), t))
        wg.flatten()
        self.cell.clear(LayerDevRecN)
        self.cell.clear(LayerPinRecN)
        self.cell.clear(LayerSiN)
        self.cell.clear(LayerWN)

        ##########################################################################################################################################################################
        # draw N tapers
        for n_ch in range(int(num_channels)):
            # draw the taper
            taper_pts = [
                Point(0, -w_waveguide / 2 + pitch * n_ch),
                Point(0, w_waveguide / 2 + pitch * n_ch),
                Point(offset + l_taper + Lw3, w_waveguide / 2 + pitch * n_ch),
                Point(offset + l_taper + Lw3, -w_waveguide / 2 + pitch * n_ch),
            ]

            # place the taper
            shapes(LayerBBN).insert(Polygon(taper_pts))

            # draw and place pin on the waveguide:
            x = offset + l_taper + Lw3
            t = Trans(Trans.R0, x, pitch * n_ch)
            pin = Path(
                [Point(-pin_length / 2, 0), Point(pin_length / 2, 0)], w_waveguide
            )
            pin_t = pin.transformed(t)
            shapes(LayerPinRecN).insert(pin_t)
            text = Text(f"opt_{n_ch+1}", t)
            shape = shapes(LayerPinRecN).insert(text)
            shape.text_size = 4 / dbu

            # draw fibre target circle
            align_circle = circle(offset, pitch * n_ch, 2 / dbu)

            # place fibre target circle
            shapes(LayerFbrTgtN).insert(Polygon(align_circle))

        # draw devrec box
        devrec_pts = [
            Point(0, pitch * n_ch + 20 / dbu),
            Point(x, pitch * n_ch + 20 / dbu),
            Point(x, -pitch * 2 - 5 / dbu),
            Point(0, -pitch * 2 - 5 / dbu),
        ]

        # place devrec box
        shapes(LayerDevRecN).insert(Polygon(devrec_pts))

        # edge of chip text
        t = Trans(Trans.R0, 0, -110 / dbu)
        text = Text("<- Edge of chip", t)
        shape = shapes(LayerTEXTN).insert(text)
        shape.text_size = 6 / dbu

        # BB description
        t = Trans(Trans.R0, 0, -95 / dbu)
        text = Text(
            "  Number of Channel(s): "
            + str(num_channels)
            + "\n  Center Wavelength: "
            + str(wavelength)
            + " nm",
            t,
        )
        shape = shapes(LayerTEXTN).insert(text)
        shape.text_size = 4 / dbu

        # BB description
        t = Trans(Trans.R0, 20 / dbu, -200 / dbu)
        text = Text("    PWB reference  ->\nloopback structure", t)
        shape = shapes(LayerTEXTN).insert(text)
        shape.text_size = 4 / dbu

        # draw DP BB logo
        import os

        dir_path = os.path.normpath(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "../../gds/EBeam_Dream/"
            )
        )
        filename = os.path.join(dir_path, "DP_Edge_coupler_for_PWB_BB_logo.gds")
        tech_name = "EBeam"
        ly2 = pya.Layout()
        ly2.read(filename)
        top_cell = ly2.top_cell()
        top_cell.flatten(True)
        # top_cell.Trans(Trans.R0,0,-192)
        # self.cell.copy_shapes(top_cell)
        self.cell.copy_shapes(top_cell)

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "ebeam_dream_PWB_edge_couplers_BB"



# This PCell instantiates two ebeam_dc_halfring_straight and adds a heater layer
# Mustafa Hammood, 2023

import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class ebeam_ring_heater(pya.PCellDeclarationHelper):
    """
    The PCell uses two ebeam_dc_halfring_straight PCells and adds a heater.
    """

    def __init__(self):
        # Important: initialize the super class
        super(ebeam_ring_heater, self).__init__()
        self.technology_name = "EBeam"
        TECHNOLOGY = get_technology_by_name(self.technology_name)

        # declare the parameters
        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param("r", self.TypeDouble, "Radius", default=10)
        self.param("w", self.TypeDouble, "Waveguide Width", default=0.5)
        self.param("g", self.TypeDouble, "Gap", default=0.1)
        self.param("Lc", self.TypeDouble, "Coupler Length", default=0.0)
        self.param(
            "orthogonal_identifier",
            self.TypeInt,
            "Orthogonal identifier (1=TE, 2=TM)",
            default=1,
        )
        self.param("heater_width", self.TypeDouble, "Metal heater width", default=3)
        self.param("metal_width", self.TypeDouble, "Metal contact width", default=5)
        self.param(
            "l_heater", self.TypeLayer, "Heater Layer", default=TECHNOLOGY["M1_heater"]
        )
        self.param(
            "l_metal",
            self.TypeLayer,
            "Metal contact Layer",
            default=TECHNOLOGY["M2_router"],
        )
        self.param(
            "pinrecm",
            self.TypeLayer,
            "PinRecM Layer (metal)",
            default=TECHNOLOGY["PinRecM"],
        )
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return (
            "ebeam_ring_heater(R="
            + ("%.3f" % self.r)
            + ",g="
            + ("%g" % (1000 * self.g))
            + ",Lc="
            + ("%g" % (1000 * self.Lc))
            + ",orthogonal_identifier="
            + ("%s" % self.orthogonal_identifier)
            + ")"
        )

    def can_create_from_shape_impl(self):
        return False

    def produce_impl(self):
        # This is the main part of the implementation: create the layout
        from SiEPIC.utils import arc
        from SiEPIC.extend import to_itype
        from SiEPIC.utils.layout import make_pin

        dbu = self.layout.dbu
        center = to_itype(self.r * 2 + self.g * 2 + self.w * 2, self.layout.dbu) / 2
        mh_layer = self.layout.layer(self.l_heater)
        ml_layer = self.layout.layer(self.l_metal)
        LayerPinRecM = self.layout.layer(self.pinrecm)
        m_radius = to_itype(self.r, dbu)
        mh_width = to_itype(self.heater_width, dbu)
        ml_width = to_itype(self.metal_width, dbu)

        # add heater
        poly = pya.Polygon(arc(m_radius + mh_width / 2, 0, 360))
        hole = pya.Polygon(arc(m_radius - mh_width / 2, 0, 360))
        poly.insert_hole(hole.get_points())
        self.cell.shapes(mh_layer).insert(
            poly.transform(pya.Trans(pya.Trans.R0, 0, center))
        )
        # add contacts
        m_length = to_itype(3, dbu)
        pts = [
            Point(m_radius, center - ml_width / 2),
            Point(m_radius, center + ml_width / 2),
            Point(m_radius + m_length, center + ml_width / 2),
            Point(m_radius + m_length, center - ml_width / 2),
        ]
        self.cell.shapes(ml_layer).insert(Polygon(pts))
        self.cell.shapes(mh_layer).insert(Polygon(pts))

        pts = [
            Point(-m_radius, center - ml_width / 2),
            Point(-m_radius, center + ml_width / 2),
            Point(-m_radius - m_length, center + ml_width / 2),
            Point(-m_radius - m_length, center - ml_width / 2),
        ]
        self.cell.shapes(ml_layer).insert(Polygon(pts))
        self.cell.shapes(mh_layer).insert(Polygon(pts))

        # add rings
        # print(self.layout.technology_name)  # the technology is not assigned in a PCell, for some reason
        self.layout.technology_name = self.technology_name

        pcell = self.layout.create_cell(
            "ebeam_dc_halfring_straight",
            self.technology_name,
            {
                "silayer": self.silayer,
                "r": self.r,
                "w": self.w,
                "g": self.g,
                "Lc": self.Lc,
                "orthogonal_identifier": self.orthogonal_identifier,
                "pinrec": self.pinrec,
                "devrec": self.devrec,
                "textl": self.textl,
            },
        )
        make_pin(
            self.cell, "elec1", [m_radius + m_length, center], ml_width, LayerPinRecM, 0
        )
        make_pin(
            self.cell,
            "elec2",
            [-m_radius - m_length, center],
            ml_width,
            LayerPinRecM,
            180,
        )

        if pcell == None:
            raise Exception(
                "problem! cannot create PCell from library: %s" % self.technology_name
            )
        inst = self.cell.insert(
            pya.CellInstArray(pcell.cell_index(), pya.Trans(pya.Trans.R0, 0, 0))
        )
        inst = self.cell.insert(
            pya.CellInstArray(
                pcell.cell_index(),
                pya.Trans(
                    pya.Trans.R180,
                    0,
                    to_itype(self.r * 2 + self.g * 2 + self.w * 2, self.layout.dbu),
                ),
            )
        )
        # print((self.r*2+self.g*2+self.w*2)/self.layout.dbu)

        # print("Done drawing the layout for - ebeam_ring_heater: %.3f-%g" % ( self.r, self.g) )



import pya
from pya import *


class ebeam_dream_splitter_1x2_te1550_BB(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the ebeam_dream_splitter_1x2_te1550_BB

    Authors: Dream Photonics
    """

    def __init__(self):
        # Important: initialize the super class
        super(ebeam_dream_splitter_1x2_te1550_BB, self).__init__()
        from SiEPIC.utils import get_technology_by_name

        self.technology_name = "EBeam"
        TECHNOLOGY = get_technology_by_name(self.technology_name)

        from SiEPIC._globals import KLAYOUT_VERSION

        if KLAYOUT_VERSION >= 28:
            # Button to launch separate window
            self.param(
                "documentation", self.TypeCallback, "Open documentation in web browser"
            )

    def callback(self, layout, name, states):
        """Callback for PCell, to launch documentation viewer
        https://www.klayout.de/doc/code/class_PCellDeclaration.html#method9
        """
        if name == "documentation":
            url = "https://github.com/SiEPIC/SiEPIC_EBeam_PDK/wiki/Component-Library-description#dream-photonics-1x2-splitter"
            import webbrowser

            webbrowser.open_new(url)

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "ebeam_dream_splitter_1x2_te1550_BB"

    def can_create_from_shape_impl(self):
        return False

    def produce_impl(self):
        # This is the main part of the implementation: create the layout

        # Load the geometries into KLayout
        import os

        dir_path = os.path.normpath(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "../../gds/EBeam_Dream/"
            )
        )
        filename = os.path.join(dir_path, "ebeam_dream_splitter_1x2_te1550_BB.gds")
        tech_name = "EBeam"
        ly2 = pya.Layout()
        ly2.read(filename)
        top_cell = ly2.top_cell()
        top_cell.flatten(True)
        self.cell.copy_shapes(top_cell)



"""
Create a layout with band bends
uses: waveguide bend cutback structure

- this can be executed directly, or loaded by the technology library
- Waveguide types are loaded from the XML file
- Pitch of the waveguides is determined by the DevRec layer
- When executed directly, the unit test instantiates all the waveguide types

PCell options:
 - waveguide type
 - ports_opposite = True: input on left, output on right
                  = False: both i/o on left
by Lukas Chrostowski, 2024
"""

import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class ebeam_test_bends(pya.PCellDeclarationHelper):
    def __init__(self):
        # Important: initialize the super class
        self.cellName = self.__class__.__name__
        super(eval(self.cellName), self).__init__()

        self.technology_name = "EBeam"

        from SiEPIC.utils import load_Waveguides_by_Tech

        self.TECHNOLOGY = get_technology_by_name(self.technology_name)

        # Load all waveguides
        waveguide_types = load_Waveguides_by_Tech(self.technology_name)

        # Make a copy so that we don't modify the original definitions
        import copy

        self.waveguide_types = copy.deepcopy(waveguide_types)
        """
        # Debugging:
        print(id(waveguide_types))
        print(id(self.waveguide_types))
        """

        # downselect to non-compound types only.
        self.waveguide_types = [
            t for t in self.waveguide_types if "compound_waveguide" not in t
        ]

        # declare the parameters
        p = self.param(
            "waveguide_type",
            self.TypeList,
            "Waveguide Type",
            default=self.waveguide_types[0]["name"],
        )
        for wa in self.waveguide_types:
            p.add_choice(wa["name"], wa["name"])
        self.param(
            "override_width", self.TypeDouble, "Override waveguide width", default=0
        )
        self.param(
            "override_radius", self.TypeDouble, "Override waveguide radius", default=0
        )
        self.param(
            "override_bezier",
            self.TypeDouble,
            "Override waveguide Bezier parameter",
            default=0,
        )
        self.param("columns", self.TypeInt, "Number of columns", default=5)
        self.param("rows", self.TypeInt, "Number of double rows", default=5)
        #        self.param("tot_bends", self.TypeInt, "Total number of bends", default = 2, readonly=True)
        self.param(
            "tot_bends", self.TypeInt, "Total number of bends", default=2, readonly=True
        )

    def coerce_parameters_impl(self):
        #        if 'tot_bends1' in dir(self):
        #            self.tot_bends = self.tot_bends1 # self.columns * self.rows
        self.tot_bends = self.columns * 2 * 4 * self.rows + 2

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "%s_%s_%s_override:%s_%s_%s" % (
            self.cellName,
            self.tot_bends,
            self.waveguide_type.replace(" ", "_"),
            self.override_width,
            self.override_radius,
            round(self.override_bezier, 3),
        )

    def get_waveguide_parameters(self):
        """
        Get the Waveguide Type from the PCell, and update waveguide parameters
        """
        from SiEPIC.utils import load_Waveguides_by_Tech

        # Load all waveguides
        waveguide_types = load_Waveguides_by_Tech(self.technology_name)
        # Make a copy so that we don't modify the original definitions
        import copy

        self.waveguide_types = copy.deepcopy(waveguide_types)

        self.waveguide_params = [
            t for t in self.waveguide_types if t["name"] == self.waveguide_type
        ]
        if not self.waveguide_params:
            raise Exception("Waveguides.XML not correctly configured")
        self.waveguide_params = self.waveguide_params[0]

        wg_params = self.waveguide_params

        if self.override_width > 0:
            delta_width = self.override_width - float(self.waveguide_params["width"])
            self.waveguide_params["width"] = str(self.override_width)
        else:
            delta_width = 0
        # Adjust all the width parameters
        for wg in wg_params["component"]:
            wg["width"] = str(float(wg["width"]) + delta_width)

        if self.override_radius > 0:
            self.waveguide_params["radius"] = str(self.override_radius)
        if self.override_bezier > 0:
            self.waveguide_params["bezier"] = str(self.override_bezier)

        # DevRec width
        if "DevRec" not in [wg["layer"] for wg in wg_params["component"]]:
            from SiEPIC import _globals

            self.devrec = (
                max([float(wg["width"]) for wg in wg_params["component"]])
                + _globals.WG_DEVREC_SPACE * 2
            )
        else:
            self.devrec = float(
                [f for f in wg_params["component"] if f["layer"] == "DevRec"][0][
                    "width"
                ]
            )

        # Radius
        self.radius = float(wg_params["radius"])

        # Waveguide pin width
        self.wg_width = float(wg_params["width"])

        """
        # Debugging:
        from SiEPIC.utils import load_Waveguides_by_Tech
        print(load_Waveguides_by_Tech(self.technology_name)[0] )
        print(id(load_Waveguides_by_Tech(self.technology_name)[0] ))
        print(self.waveguide_types[0])
        print(id(self.waveguide_types[0]))
        print(wg_params)     
        """

    def produce_impl(self):
        self.get_waveguide_parameters()
        devrec = self.devrec
        radius = self.radius

        if 0:
            # Spiral

            # SBend offset in the middle of the spiral
            if "sbends" in self.waveguide_params:
                # Use the Bezier S-Bend (more space efficient)
                # if sbends=True in Waveguide.xml
                offset = radius - devrec / 2
                extra = 0
            else:
                # Use 2 x 90 degree bends
                offset = radius
                extra = devrec

            # Ensure the length is sufficient
            self.minlength = 2 * radius
            length0 = max(self.minlength, self.length)

            # min loops
            self.loops = max(self.loops, 1)

            # spiral points
            points = [
                DPoint(-length0, offset),
                DPoint(0.0, offset),
                DPoint(0.0, -offset),
                DPoint(length0, -offset),
            ]
            for i in range(1, self.loops * 2, 2):
                points.insert(
                    0,
                    DPoint(
                        -length0 - devrec * (i - 1),
                        offset - radius * 2 - devrec * (i - 1) - extra,
                    ),
                )
                points.insert(
                    0,
                    DPoint(
                        length0 + devrec * i,
                        offset - radius * 2 - devrec * (i - 1) - extra,
                    ),
                )
                points.insert(
                    0,
                    DPoint(
                        length0 + devrec * i, -offset + radius * 2 + devrec * i + extra
                    ),
                )
                points.insert(
                    0,
                    DPoint(
                        -length0 - devrec * (i + 1),
                        -offset + radius * 2 + devrec * i + extra,
                    ),
                )
                points.append(
                    DPoint(
                        length0 + devrec * (i - 1),
                        radius * 2 - offset + devrec * (i - 1) + extra,
                    )
                )
                points.append(
                    DPoint(
                        -length0 - devrec * i,
                        radius * 2 - offset + devrec * (i - 1) + extra,
                    )
                )
                points.append(
                    DPoint(
                        -length0 - devrec * i, -radius * 2 + offset - devrec * i - extra
                    )
                )
                points.append(
                    DPoint(
                        length0 + devrec * (i + 1),
                        -radius * 2 + offset - devrec * i - extra,
                    )
                )
            if not self.ports_opposite:
                points.append(
                    DPoint(
                        length0 + devrec * (i + 1),
                        radius * 2 - offset + devrec * (i + 1) + extra,
                    )
                )
                points.append(
                    DPoint(
                        -length0 - devrec * (i + 1),
                        radius * 2 - offset + devrec * (i + 1) + extra,
                    )
                )
            points.pop(0)
            points.insert(
                0,
                DPoint(
                    -length0 - devrec * (i + 1),
                    -offset + radius * 2 + devrec * i + extra,
                ),
            )

        else:
            # bend test structure
            points = [DPoint(0, 0)]
            for j in range(0, self.rows):
                dy = radius * 4 * j
                for i in range(1, self.columns * 2 + 1, 2):
                    points.append(DPoint(radius * 2 * (i), dy))
                    points.append(DPoint(radius * 2 * (i), radius * 2 + dy))
                    points.append(DPoint(radius * 2 * (i + 1), radius * 2 + dy))
                    points.append(DPoint(radius * 2 * (i + 1), dy))
                points.append(DPoint(radius * 2 * (i + 2), dy))
                points.append(DPoint(radius * 2 * (i + 2), radius * 2 + dy))
                for i in range(self.columns * 2 - 1, 0, -2):
                    points.append(DPoint(radius * 2 * (i + 1), radius * 2 + dy))
                    points.append(DPoint(radius * 2 * (i + 1), radius * 4 + dy))
                    points.append(DPoint(radius * 2 * (i), radius * 4 + dy))
                    points.append(DPoint(radius * 2 * (i), radius * 2 + dy))
                points.pop(-1)
                points.pop(-1)
            #            points.append(DPoint(radius*2*(i+2),radius*2*self.rows))
            points.append(DPoint(0, radius * 4 * self.rows))
            self.tot_bends1 = len(points)
            print("Number of bends: %s" % self.tot_bends)

        # Create a path and waveguide
        path = DPath(points, 0.5)
        """
        # Debuggin, to make sure there are no duplicate vertices
        print(path.to_s())
        path = path.unique_points()
        print(path.to_s())
        """

        if 0:
            # This inserts a waveguide where the parameters are fixed from Waveguide.XML
            # using the Waveguide PCell
            self.layout.technology_name = (
                self.technology_name
            )  # required otherwise "create_cell" doesn't load
            pcell = self.layout.create_cell(
                "Waveguide",
                self.technology_name,
                {"path": path, "waveguide_type": self.waveguide_type},
            )
            t = Trans(Trans.R0, 0, 0)
            self.cell.insert(CellInstArray(pcell.cell_index(), t))
        else:
            # This inserts a waveguide where the parameters can be modified
            # using the function layout_waveguide3.
            from SiEPIC.utils.layout import layout_waveguide3

            # Make sure the technology name is associated with the layout
            #  PCells don't seem to know to whom they belong!
            if self.layout.technology_name == "":
                self.layout.technology_name = self.technology_name
            self.waveguide_params["waveguide_type"] = self.waveguide_type
            ipoints = [p.to_itype(self.layout.dbu) for p in points]

            waveguide_length = layout_waveguide3(
                self.cell, ipoints, self.waveguide_params, debug=False
            )
            print("done layout_waveguide3 test_bends")


# Save the path, used for loading WAVEGUIDES.XML
import os

filepath = os.path.dirname(os.path.realpath(__file__))


class test_lib(Library):
    def __init__(self):
        tech = "EBeam"
        library = tech + "test_lib"
        self.technology = tech
        self.layout().register_pcell("ebeam_test_bends", ebeam_test_bends())
        self.register(library)


if __name__ == "__main__":
    print("Test layout for: test bends")

    from SiEPIC.utils.layout import new_layout
    from SiEPIC.scripts import zoom_out

    # load the test library, and technology
    t = test_lib()
    tech = t.technology

    # Create a new layout for the chip floor plan
    topcell, ly = new_layout(tech, "test", GUI=True, overwrite=True)
    # floorplan(topcell, 100e3, 100e3)

    # instantiate the cell
    library = tech + "test_lib"

    # Create test structures for all the types of waveguides
    from SiEPIC.utils import load_Waveguides_by_Tech

    waveguide_types = load_Waveguides_by_Tech(tech)
    xmax = 0
    ports_opposite = False
    flatten = False
    y = 0
    x = xmax
    if 1:
        # only run the first type
        waveguide_types = [waveguide_types[0]]
    for wg in waveguide_types:
        pcell = ly.create_cell(
            "ebeam_test_bends",
            library,
            {
                "waveguide_type": wg["name"],
                "length": 100,
                "loops": 1,
                "flatten": flatten,
                "ports_opposite": ports_opposite,
            },
        )
        t = Trans(Trans.R0, x, y - pcell.bbox().bottom)
        inst = topcell.insert(CellInstArray(pcell.cell_index(), t))
        y += pcell.bbox().height() + 2000
        xmax = max(xmax, x + inst.bbox().width())

    zoom_out(topcell)



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class SWG_MultiBox_Ring(pya.PCellDeclarationHelper):
    def __init__(self):
        # Important: initialize the super class
        super(SWG_MultiBox_Ring, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("pitch", self.TypeDouble, "Period [um]", default=0.24)
        self.param("w", self.TypeDouble, "Waveguide Width [um]", default=0.18)
        self.param(
            "r", self.TypeDouble, "Radius [um]", default=35
        )  # minimum radius is 5 um in TE
        self.param("ff", self.TypeDouble, "Duty Cycle", default=0.75)
        self.param("angle", self.TypeDouble, "Angle", default=360)
        self.param("gap", self.TypeDouble, "Row Gap [um]", default=0.06)
        self.param("gap2", self.TypeDouble, "Bus Gap [um]", default=0.06)
        self.param("row", self.TypeDouble, "Number of Rows", default=5)
        self.param("busL", self.TypeDouble, "Bus Length [um]", default=10)
        self.param("taperL", self.TypeDouble, "Taper Length [um]", default=25)
        # self.param("doublebus", self.TypeBoolean, "Bus Type (Single=False, Double=True)", default = False)
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        #    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))
        self.param(
            "oxideopen",
            self.TypeLayer,
            "Oxide Open Layer",
            default=TECHNOLOGY["Oxide open (to BOX)"],
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "SWG_multibox_%.3f" % (self.pitch)

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        debug = False
        from SiEPIC.extend import to_itype
        import math
        from pya import DPolygon

        pi = math.pi
        # This is the main part of the implementation: create the layout

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes
        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        # variables
        pitch = self.pitch
        w = self.w
        r = self.r
        ff = self.ff
        angle = self.angle
        gap = self.gap
        gap2 = self.gap2
        row = self.row
        busL = self.busL
        taperL = self.taperL
        # doublebus = self.doublebus
        # print(doublebus)

        if r - w / 2 <= 0:
            r = 5
            if debug:
                print("invalid radius, set r to default: 5")
        if row <= 1:
            row = 1
            if debug:
                print("invalid number of rows, set row to default: 1")
        if busL <= 10:
            busL = taperL * 2 + 10
            if debug:
                print(
                    "invalid length of SWG bus waveguide, set length at least 2 times larger than taper length and pluse 10 um"
                )
        # Calculate number of segments
        s1 = pitch * ff  # silicon
        s2 = pitch - s1  # gap

        # draw oxide open
        shapes(ly.layer(self.oxideopen)).insert(
            pya.Box(
                DPoint(-to_itype(r + 5, dbu), -to_itype(r + 5, dbu)),
                DPoint(to_itype(r + 5, dbu), to_itype(r + 5, dbu)),
            )
        )

        # Draw the Multi-box Ring
        for i in range(0, int(row)):  # draw different radius SWG rings
            # calculate best radius
            # const = math.ceil(2*pi*r/(s1+s2))
            const = math.floor(2 * pi * r / (s1 + s2))
            # if doesn't divide evenly, replace r with best possible r
            if (2 * pi * r) % (s1 + s2) != 0:
                r = const * (s1 + s2) / (2 * pi)
                if debug:
                    print("r adjusted to " + str(r) + "um to fit periods perfectly.")

            theta1 = math.atan(s1 / r)
            theta2 = math.atan(s2 / r)
            nSeg = int(
                math.floor(angle / (math.degrees(theta1) + math.degrees(theta2)))
            )  # how many segments to have
            si_first = True  # for alternating between silicon and gap
            j = 0  # index of how many silicon thetas
            jj = 0  # index of how many gap thetas
            ORDER = True  # ordering of the coordinates for polygon drawing

            xo = [(r - w / 2) * math.cos(0)]
            yo = [(r - w / 2) * math.sin(0)]
            xo.append((r + w / 2) * math.cos(0))
            yo.append((r + w / 2) * math.sin(0))

            for i in range(0, nSeg * 2):
                if si_first:
                    j = j + 1
                    si_first = not (si_first)
                else:
                    jj = jj + 1
                    si_first = not (si_first)

                if ORDER:
                    xo.append((r + w / 2) * math.cos(j * theta1 + jj * theta2))
                    yo.append((r + w / 2) * math.sin(j * theta1 + jj * theta2))
                    xo.append((r - w / 2) * math.cos(j * theta1 + jj * theta2))
                    yo.append((r - w / 2) * math.sin(j * theta1 + jj * theta2))
                    ORDER = not (ORDER)
                else:
                    xo.append((r - w / 2) * math.cos(j * theta1 + jj * theta2))
                    yo.append((r - w / 2) * math.sin(j * theta1 + jj * theta2))
                    xo.append((r + w / 2) * math.cos(j * theta1 + jj * theta2))
                    yo.append((r + w / 2) * math.sin(j * theta1 + jj * theta2))
                    ORDER = not (ORDER)

                if len(xo) == 4:
                    dpts = [pya.DPoint(xo[i], yo[i]) for i in range(len(xo))]
                    dpolygon = DPolygon(dpts)
                    element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                    shapes(LayerSiN).insert(element)
                    xo = []
                    yo = []
            r = r - (w + gap)
        # Draw the Bus Waveguides
        # Draw Bus WG
        # to go back to the initial point for Multi-box drawing, ger the first adjusted r value
        r = self.r
        const = math.floor(2 * pi * r / (s1 + s2))
        # if doesn't divide evenly, replace r with best possible r
        if (2 * pi * r) % (s1 + s2) != 0:
            r = const * (s1 + s2) / (2 * pi)

        # calulate ideal length of bus
        bus_length = busL
        # bus_length = self.cell.bbox().height()*dbu +pitch*2
        constant = math.ceil(bus_length / (s1 + s2))
        if bus_length % (s1 + s2) != 0:
            bus_length = constant * (s1 + s2)

        for ii in range(0, int(row)):
            xo = [
                (r + w / 2 + gap2 + ii * (w + gap)),
                (r + w / 2 + gap2 + w + ii * (w + gap)),
                (r + w / 2 + gap2 + w + ii * (w + gap)),
                (r + w / 2 + gap2 + ii * (w + gap)),
            ]
            yo = [0, 0, s1, s1]
            dpts = [pya.DPoint(xo[i], yo[i]) for i in range(len(xo))]
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
            shapes(LayerSiN).insert(element)

            # draw the bus waveguide
            for i in range(0, int(math.ceil((constant) / 2))):
                yu = [yo[j] + i * pitch for j in range(len(yo))]
                yd = [yo[j] - i * pitch for j in range(len(yo))]

                dpts = [pya.DPoint(xo[i], yu[i]) for i in range(len(xo))]
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                shapes(LayerSiN).insert(element)
                dpts = [pya.DPoint(xo[i], yd[i]) for i in range(len(xo))]
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                shapes(LayerSiN).insert(element)

            # draw the tapers from waveguide to SWG
            xtu = [
                ((r + w / 2 + gap2 + ii * (w + gap)) + (w - 0.06) / 2),
                ((r + w / 2 + gap2 + ii * (w + gap)) + (w - 0.06) / 2 + 0.06),
                ((r + w / 2 + gap2 + ii * (w + gap)) + w + gap / 2),
                ((r + w / 2 + gap2 + ii * (w + gap)) - gap / 2),
            ]
            ytu = [(yu[3] - taperL), (yu[3] - taperL), (yu[3]), (yu[3])]
            dpts = [pya.DPoint(xtu[i], ytu[i]) for i in range(len(xtu))]
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
            shapes(LayerSiN).insert(element)

            ytd = [(yd[1] + taperL), (yd[1] + taperL), (yd[1]), (yd[1])]
            dpts = [pya.DPoint(xtu[i], ytd[i]) for i in range(len(xtu))]
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
            shapes(LayerSiN).insert(element)

        # taper connections from multi-tapers to std WG
        xTu = [
            (r + w / 2 + gap2 - gap / 2),
            (r + w / 2 + gap2 - gap / 2 + row * (gap + w)),
            (r + w / 2 + gap2 + row * (gap + w) / 2 + 0.25),
            (r + w / 2 + gap2 + row * (gap + w) / 2 - 0.25),
        ]
        yTu = [(yu[3]), (yu[3]), (yu[3] + taperL), (yu[3] + taperL)]
        dpts = [pya.DPoint(xTu[i], yTu[i]) for i in range(len(xTu))]
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
        shapes(LayerSiN).insert(element)

        yTd = [(yd[1]), (yd[1]), (yd[1] - taperL), (yd[1] - taperL)]
        dpts = [pya.DPoint(xTu[i], yTd[i]) for i in range(len(xTu))]
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
        shapes(LayerSiN).insert(element)

        # add std WG
        if 2 * r >= busL + 2 * taperL:
            xstdu = [
                (r + w / 2 + gap2 + row * (gap + w) / 2 - 0.25),
                (r + w / 2 + gap2 + row * (gap + w) / 2 + 0.25),
                (r + w / 2 + gap2 + row * (gap + w) / 2 + 0.25),
                (r + w / 2 + gap2 + row * (gap + w) / 2 - 0.25),
            ]
            ystdu = [
                (yu[3] + taperL),
                (yu[3] + taperL),
                (yu[3] + r - busL / 2 + 1),
                (yu[3] + r - busL / 2 + 1),
            ]
            dpts = [pya.DPoint(xstdu[i], ystdu[i]) for i in range(len(xstdu))]
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
            shapes(LayerSiN).insert(element)

            ystdd = [
                (yd[1] - taperL),
                (yd[1] - taperL),
                (yd[1] - r + busL / 2 - 1),
                (yd[1] - r + busL / 2 - 1),
            ]
            dpts = [pya.DPoint(xstdu[i], ystdd[i]) for i in range(len(xstdu))]
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
            shapes(LayerSiN).insert(element)
        else:
            xstdu = xTu
            ystdu = yTu
            ystdd = yTd

        # DEV BOX
        half_l = (
            self.cell.bbox().width()
            - row * (gap + w) / dbu
            - gap2 / dbu
            + (gap / 2) / dbu
        ) / 2
        half_r = self.cell.bbox().width() - half_l
        dev = Box(-half_l, ystdd[-1] / dbu - 1, half_r, ystdu[-1] / dbu + 1)
        shapes(LayerDevRecN).insert(dev)
        dev_width = self.cell.bbox().width() / 2

        if 2 * r >= busL + 2 * taperL:
            dev_up = (
                yu[-1] / dbu
                + taperL / dbu
                + (yu[3] + r - busL / 2 + 1 - (yu[3] + taperL)) / dbu
            )
            dev_down = (
                yd[-4] / dbu
                - taperL / dbu
                - (yu[3] + r - busL / 2 + 1 - (yu[3] + taperL)) / dbu
            )
        else:
            dev_up = yu[-1] / dbu + taperL / dbu
            dev_down = yd[-4] / dbu - taperL / dbu

        # Create the pins on the waveguides, as short paths:
        from SiEPIC._globals import PIN_LENGTH as pin_length

        w = to_itype(self.w, dbu)
        gap = to_itype(self.gap, dbu)
        bus_length = to_itype(bus_length / 2, dbu)

        # Pin1
        # t = Trans(Trans.R0, dev_width-w/2,dev_up)
        t = Trans(Trans.R0, xstdu[-1] / dbu + 251, dev_up + 1)
        pin = Path(
            [Point(0, -pin_length / 2), Point(0, pin_length / 2)], 500
        )  # 500 is width of std WG
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("opt1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu
        if debug:
            print(dev_up)
            print(dev_up * dbu)
        # Pin2
        t = Trans(Trans.R0, xstdu[-1] / dbu + 251, dev_down - 1)
        pin = Path([Point(0, pin_length / 2), Point(0, -pin_length / 2)], 500)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("opt2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name
from SiEPIC.utils.layout import make_pin


class SWG_Ring(pya.PCellDeclarationHelper):
    """
    Author: Ben Cohen (UBC), Kithmin Wickremasinghe (UBC)
    Date: 2024-07-22
    This Pcell creates a sub-wavelengh grating (SWG) ring resonator

    Parameters:
      pitch  - grating period [um]
      w      - waveguide width [um]
      r      - ring radii [um]
      ff     - grating duty cycle (%)
      angle  - angle of the resonator (360 = full ring)
      gap2   - coupling gap [um]
      busL   - bus waveguide length [um]
      taperL - SWG <-> Strip Taper
    """

    def __init__(self):
        # Important: initialize the super class
        super(SWG_Ring, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("pitch", self.TypeDouble, "Period [um]", default=0.25)
        self.param("w", self.TypeDouble, "Waveguide Width [um]", default=0.5)
        self.param(
            "r", self.TypeDouble, "Radius [um]", default=30
        )  # minimum radius is 5 um in TE
        self.param("ff", self.TypeDouble, "Duty Cycle", default=0.5)
        self.param("angle", self.TypeDouble, "Angle", default=360)
        self.param("gap2", self.TypeDouble, "Bus Gap [um]", default=0.06)
        self.param("busL", self.TypeDouble, "Bus Length [um]", default=100)
        self.param("taperL", self.TypeDouble, "Taper Length [um]", default=30)

        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param(
            "oxideopen",
            self.TypeLayer,
            "Oxide Open Layer",
            default=TECHNOLOGY["Oxide open (to BOX)"],
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "SWG_multibox_%.3f" % (self.pitch)

    def produce_impl(self):
        debug = False
        from SiEPIC.extend import to_itype
        import math
        from pya import DPolygon

        pi = math.pi

        ###################### Instantiations + Checking

        # Fetch the Parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes
        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        # Instantiate the Pcell Parameters
        pitch = self.pitch  # SWG Period [um]
        w = self.w  # Waveguide Width [um]
        r = self.r  # Ring Radii [um]
        ff = self.ff  # SWG Duty Cycle [%]
        angle = self.angle  # Angle of the ring
        gap2 = self.gap2  # Coupling Gap [um]
        busL = self.busL  # Bus Length [um]
        taperL = self.taperL  # Taper Length [um]

        # Minimize Ring Radii to 5 um
        if r - w / 2 <= 0:
            r = 5
            if debug:
                print("invalid radius, set r to default: 5")

        # Bus waveguide should be longer than 2x*taperL + 10 um for coupling
        if busL <= 10:
            busL = taperL * 2 + 10
            if debug:
                print(
                    "invalid length of SWG bus waveguide, set length at least 2 times larger than taper length and plus 10 um"
                )

        # Calculate number of SWG segments within the resonator
        s1 = pitch * ff  # silicon
        s2 = pitch - s1  # gap

        # draw oxide open
        shapes(ly.layer(self.oxideopen)).insert(
            pya.Box(
                DPoint(-to_itype(r + 5, dbu), -to_itype(r + 5, dbu)),
                DPoint(to_itype(r + 5, dbu), to_itype(r + 5, dbu)),
            )
        )

        ###################### Draw the Multi-box Ring

        # Calculate a radius to fit all of the gratings
        const = math.floor(2 * pi * r / (s1 + s2))
        # if doesn't divide evenly, replace r with best possible r
        if (2 * pi * r) % (s1 + s2) != 0:
            r = const * (s1 + s2) / (2 * pi)
            if debug:
                print("r adjusted to " + str(r) + "um to fit periods perfectly.")

        # Draw the SWGs
        theta1 = math.atan(s1 / r)
        theta2 = math.atan(s2 / r)
        nSeg = int(
            math.floor(angle / (math.degrees(theta1) + math.degrees(theta2)))
        )  # NUmber of SWG segments

        si_first = True  # for alternating between silicon and gap
        j = 0  # Index of how many silicon thetas
        jj = 0  # Index of how many gap thetas
        ORDER = True  # Ordering of the coordinates for polygon drawing

        xo = [(r - w / 2) * math.cos(0)]
        yo = [(r - w / 2) * math.sin(0)]
        xo.append((r + w / 2) * math.cos(0))
        yo.append((r + w / 2) * math.sin(0))

        for i in range(0, nSeg * 2):
            if si_first:
                j = j + 1
                si_first = not (si_first)
            else:
                jj = jj + 1
                si_first = not (si_first)

            if ORDER:
                xo.append((r + w / 2) * math.cos(j * theta1 + jj * theta2))
                yo.append((r + w / 2) * math.sin(j * theta1 + jj * theta2))
                xo.append((r - w / 2) * math.cos(j * theta1 + jj * theta2))
                yo.append((r - w / 2) * math.sin(j * theta1 + jj * theta2))
                ORDER = not (ORDER)
            else:
                xo.append((r - w / 2) * math.cos(j * theta1 + jj * theta2))
                yo.append((r - w / 2) * math.sin(j * theta1 + jj * theta2))
                xo.append((r + w / 2) * math.cos(j * theta1 + jj * theta2))
                yo.append((r + w / 2) * math.sin(j * theta1 + jj * theta2))
                ORDER = not (ORDER)

            if len(xo) == 4:
                dpts = [pya.DPoint(xo[i], yo[i]) for i in range(len(xo))]
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                shapes(LayerSiN).insert(element)
                xo = []
                yo = []
        r = r - w

        # Draw the Bus Waveguides
        r = self.r  # to go back to the initial point for Multi-box drawing, ger the first adjusted r value
        const = math.floor(2 * pi * r / (s1 + s2))

        # If doesn't divide evenly, replace r with best possible r
        if (2 * pi * r) % (s1 + s2) != 0:
            r = const * (s1 + s2) / (2 * pi)

        # calulate ideal length of bus
        bus_length = busL
        constant = math.ceil(bus_length / (s1 + s2))
        if bus_length % (s1 + s2) != 0:
            bus_length = constant * (s1 + s2)

        xo = [
            (r + w / 2 + gap2),
            (r + w / 2 + gap2 + w),
            (r + w / 2 + gap2 + w),
            (r + w / 2 + gap2),
        ]
        yo = [0, 0, s1, s1]
        dpts = [pya.DPoint(xo[i], yo[i]) for i in range(len(xo))]
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
        shapes(LayerSiN).insert(element)

        # draw the bus waveguide
        for i in range(0, int(math.ceil((constant) / 2))):
            yu = [yo[j] + i * pitch for j in range(len(yo))]
            yd = [yo[j] - i * pitch for j in range(len(yo))]

            dpts = [pya.DPoint(xo[i], yu[i]) for i in range(len(xo))]
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
            shapes(LayerSiN).insert(element)
            dpts = [pya.DPoint(xo[i], yd[i]) for i in range(len(xo))]
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
            shapes(LayerSiN).insert(element)

        # draw the tapers from waveguide to SWG
        xtu = [
            ((r + w / 2 + gap2) + (w - 0.06) / 2),
            ((r + w / 2 + gap2) + (w - 0.06) / 2 + 0.06),
            ((r + w / 2 + gap2) + w),
            (r + w / 2 + gap2),
        ]
        ytu = [(yu[3] - taperL), (yu[3] - taperL), (yu[3]), (yu[3])]
        dpts = [pya.DPoint(xtu[i], ytu[i]) for i in range(len(xtu))]
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
        shapes(LayerSiN).insert(element)

        ytd = [(yd[1] + taperL), (yd[1] + taperL), (yd[1]), (yd[1])]
        dpts = [pya.DPoint(xtu[i], ytd[i]) for i in range(len(xtu))]
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
        shapes(LayerSiN).insert(element)

        # Bus Strip Waveguide
        xTu = [min(xtu), min(xtu), min(xtu) + w, min(xtu) + w]

        yTu = [max(ytu), round(max(ytu) + busL), round(max(ytu) + busL), max(ytu)]

        dpts = [pya.DPoint(xTu[i], yTu[i]) for i in range(len(xtu))]
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
        shapes(LayerSiN).insert(element)

        yTd = [min(ytd), round(min(ytd) - busL), round(min(ytd) - busL), min(ytd)]

        dpts = [pya.DPoint(xTu[i], yTd[i]) for i in range(len(xtu))]
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
        shapes(LayerSiN).insert(element)

        # DEV BOX
        half_l = (self.cell.bbox().width() - (w) / dbu - gap2 / dbu) / 2
        half_r = self.cell.bbox().width() - half_l
        dev = Box(-half_l, max(yTu) / dbu, half_r, min(yTd) / dbu)
        shapes(LayerDevRecN).insert(dev)
        dev_width = self.cell.bbox().width() / 2

        # Create the pins on the waveguides, as short paths:

        bus_length = to_itype(bus_length / 2, dbu)

        # Pin1
        x_pin = max(xTu) - w / 2
        y_pin = max(yTu)
        make_pin(self.cell, "opt1", [x_pin, y_pin], w, LayerPinRecN, 90)

        # t = DTrans(Trans.R0, (max(xTu) - w/2)/dbu, max(yTu)/dbu)
        # pin = Path([Point(0, -pin_length/2), Point(0, pin_length/2)], 500) #500 is width of std WG
        # pin_t = pin.transformed(t)
        # shapes(LayerPinRecN).insert(pin_t)
        # text = Text ("opt1", t)
        # shape = shapes(LayerPinRecN).insert(text)
        # shape.text_size = 0.4/dbu

        # Pin2
        x_pin = max(xTu) - w / 2
        y_pin = min(yTd)
        make_pin(self.cell, "opt2", [x_pin, y_pin], w, LayerPinRecN, 270)

        # t = DTrans(Trans.R0, (max(xTu) - w/2)/dbu, min(yTd)/dbu)
        # pin = Path([Point(0, pin_length/2), Point(0, -pin_length/2)], 500) #500 is width of std WG
        # pin_t = pin.transformed(t)
        # shapes(LayerPinRecN).insert(pin_t)
        # text = Text ("opt2", t)
        # shape = shapes(LayerPinRecN).insert(text)
        # shape.text_size = 0.4/dbu
